from abc import ABC, abstractmethod
import warnings
from typing import Union, Optional  # for python<=3.9

import numpy as np
from scipy.interpolate import interp1d

from dispersionrelations.constants import GeV, alpha_fs, E_almost_zero
from dispersionrelations.kinematics import phase_space_twobody
from dispersionrelations.integrals import DispersionIntegralRHC, integrate_gl


def sR(M_R, G_R):
    r"""
    Resonance pole location calculated from its mass and width.

    Parameters
    ----------
    M_R : float
        Mass of the particle.
    G_R : float
        Width of the particle.

    Returns
    -------
    sR : complex
        The complex pole location.

    Notes
    -----
    The pole location is calculated using :math:`s_R(M_R, \Gamma_R) = (M_R - i \Gamma_R / 2)^2`.
    """
    return (M_R - 1j * G_R / 2) ** 2



def vertex_VPP__2(s, mP, mP_2=None):
    r"""A vector :math:`\to` pseudoscalar–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mP : float
        Mass of the pseudoscalar with momentum :math:`p_1`.

    mP_2 : float
        Mass of the pseudoscalar with momentum :math:`p_2` (if not passed, a case with equal masses will be returned).

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    The vertex :math:`V(q)\to P(p_1)P(p_2)` is given by

    .. math::
        \hat{\beta}_{VPP}^{\mu}(q,p_1,p_2) \propto (p_1 - p_2)^{\mu}.

    Spin-averaged vertex function 

    .. math::
        \beta_{VPP}^2(q^2, p_1^2, p_2^2) = \frac{\lambda(q^2, p_1^2, p_2^2)}{q^2} .

    The function returns :math:`\beta_{VPP}^2(s, m_{P1}^2, m_{P2}^2)/3`,
    where :math:`3=(2l+1)` corresponds to the P-wave :math:`(l=1)`.
    """
    if mP_2 is None:
        return 1 / 3 * (s - 4 * mP**2)
    return 1 / 3 * (s - (mP + mP_2) ** 2) * (s - (mP - mP_2) ** 2) / s


def vertex_VPP(s, mP, mP_2=None):
    r"""A vector :math:`\to` pseudoscalar–pseudoscalar vertex.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mP : float
        Mass of the pseudoscalar with momentum :math:`p_1`.

    mP_2 : float
        Mass of the pseudoscalar with momentum :math:`p_2` (if not passed, a case with equal masses will be returned).

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    This returns the square root of :func:`vertex_VPP__2`, i.e. :math:`\frac{1}{\sqrt{3}}\beta_{VPP}(s, m_{P1}^2, m_{P2}^2)`.
    For most applications we need the squared version.
    """
    return np.sqrt(vertex_VPP__2(s, mP, mP_2) + 0j)


def vertex_VVP__2(s, mV, mP):
    r"""A vector :math:`\to` vector–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    The vertex :math:`V(q)\to V(k)P(p)` is given by

    .. math::
        \hat{\beta}_{VVP}^{(\lambda)\mu}(q, k, p) = \epsilon^{\mu\nu\alpha\beta} n^{(\lambda)}_\nu p_\alpha q_\beta .

    Spin-averaged vertex function 

    .. math::
        \beta_{VVP}^2(q^2, k^2, p^2) = \frac{\lambda(q^2, k^2, p^2)}{2} .

    The function returns :math:`\beta_{VVP}^2(s, m_V^2, m_P^2)/3`,
    where :math:`3=(2l+1)` corresponds to the P-wave :math:`(l=1)`.
    """
    return 1 / 6 * (s - (mV - mP) ** 2) * (s - (mV + mP) ** 2)


def vertex_VVP(s, mV, mP):
    r"""A vector :math:`\to` vector–pseudoscalar vertex.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    This returns the square root of :func:`vertex_VVP__2`, i.e. :math:`\frac{1}{\sqrt{3}}\beta_{VVP}(s, m_V^2, m_P^2)`.
    For most applications we need the squared version.
    """
    return np.sqrt(vertex_VVP__2(s, mV, mP) + 0j)


def vertex_AVP__2(s, mV, mP):
    r"""An axial vector :math:`\to` vector–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    The vertex :math:`A(q)\to V(k)P(p)` is given by

    .. math::
        \hat{\beta}_{AVP}^{(\lambda)\mu}(q, k, p) = (q \cdot k) \, n^{(\lambda)\mu} - (n^{(\lambda)} \cdot q) \, k^\mu .

    Spin-averaged vertex function 

    .. math::
        \beta_{AVP}^2(q^2, k^2, p^2) = 3q^2k^2 + \frac{\lambda(q^2, k^2, p^2)}{2} .

    The function returns :math:`\beta_{AVP}^2(s, m_V^2, m_P^2)/3`.
    """
    return 1 / 6 * (2 * s * mV**2 + (s + mV**2 - mP**2) ** 2)


def vertex_AVP(s, mV, mP):
    r"""An axial vector :math:`\to` vector–pseudoscalar vertex.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    This returns the square root of :func:`vertex_AVP__2`, i.e. :math:`\frac{1}{\sqrt{3}}\beta_{AVP}(s, m_V^2, m_P^2)`.
    For most applications we need the squared version.
    """
    return np.sqrt(vertex_AVP__2(s, mV, mP) + 0j)


def vertex_VAP__2(s, mA, mP):
    r"""A vector :math:`\to` axial vector–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the axial vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    The vertex :math:`V(q)\to A(k)P(p)` is given by

    .. math::
        \hat{\beta}_{VAP}^{(\lambda)\mu}(q, k, p) = (q \cdot k) \, n^{(\lambda)\mu} - (n^{(\lambda)} \cdot q) \, k^\mu .

    Spin-averaged vertex function 

    .. math::
        \beta_{VAP}^2(q^2, k^2, p^2) = 3q^2k^2 + \frac{\lambda(q^2, k^2, p^2)}{2} .

    The function returns :math:`\beta_{VAP}^2(s, m_A^2, m_P^2)/3`.
    """
    return 1 / 6 * (2 * s * mA**2 + (s + mA**2 - mP**2) ** 2)


def vertex_VAP(s, mA, mP):
    r"""A vector :math:`\to` axial vector–pseudoscalar vertex.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the axial vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.

    Notes
    -----
    This returns the square root of :func:`vertex_VAP__2`, i.e. :math:`\frac{1}{\sqrt{3}}\beta_{VAP}(s, m_A^2, m_P^2)`.
    For most applications we need the squared version.
    """
    return np.sqrt(vertex_VAP__2(s, mA, mP) + 0j)


def vertex_VAP_T1__2(s, mA, mP):
    r"""A vector :math:`\to` axial vector–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the axial vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.
    """
    return s / 3


def vertex_VAP_T2__2(s, mA, mP):
    r"""A vector :math:`\to` axial vector–pseudoscalar vertex squared.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).

    mV : float
        Mass of the axial vector with momentum :math:`k`.

    mP : float
        Mass of the pseudoscalar with momentum :math:`p`.

    Returns
    -------
    b : array_like
        The same shape as input `s`.
    """
    return 1 / 6 + s * 0


def taming_BlattWeisskopf(
    s, sthr: float, sB: float, l: int = 1, normalize_at_sB: bool = False
):
    r"""
    Blatt–Weisskopf taming factors.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).
    sthr : float
        Threshold of the relevant channel.
    sB : float
        Range parameter.
    l : int
        Partial wave.
    normalize_at_sB : boolean
        If set to `True`, :math:`B(s_B) = 1`.

    Returns
    -------
    B : array_like
        The same shape as input `s`.

    Raises
    ------
    NotImplementedError
        The function is only implemented for the partial waves :math:`l\in\{0,1,2,3,4\}`.
        Any other input triggers an error.

    Notes
    -----
    The first few Blatt–Weisskopf taming factors are given by :cite:`Chung:1995dx`

    .. math::

        \begin{align}
        B_0(s) &= 1, \\
        B_1(s) &= 1 / (x + 1), \\
        B_2(s) &= 1 / ((x-3)^2 + 9x), \\
        \end{align}

    where :math:`x=(s-s_{thr}) / (s_B - s_{thr})`.
    """
    if l == 0:
        return 1
    elif l == 1:
        x = (s - sthr) / (sB - sthr)
        numerator = 2 * x if normalize_at_sB else 1
        return numerator / (1 + x)
    elif l == 2:
        x = (s - sthr) / (sB - sthr)
        numerator = 13 * x**2 if normalize_at_sB else 1
        return numerator / ((x - 3) ** 2 + 9 * x)
    elif l == 3:
        x = (s - sthr) / (sB - sthr)
        numerator = 277 * x**3 if normalize_at_sB else 1
        return numerator / (x * (x - 15) ** 2 + 9 * (2 * x - 5))
    elif l == 4:
        x = (s - sthr) / (sB - sthr)
        numerator = 12746 * x**4 if normalize_at_sB else 1
        return numerator / ((x**2 - 45 * x + 105) ** 2 + 25 * x * (2 * x - 21) ** 2)
    else:
        raise NotImplementedError(
            f"Blatt-Weisskopf factor for l={l} has not been implemented."
        )


def taming_spacelike_pole(s, sB: float = (2 * GeV) ** 2, n: int = 1):
    r"""
    A simple decaying function, used as a taming factor.

    Parameters
    ----------
    s : array_like
        Kinematic variable (can be complex).
    sB : float
        Range parameter.
    n : int
        Degree of decay.

    Returns
    -------
    B : array_like
        The same shape as input `s`.

    Notes
    -----
    The function is defined as

    .. math:: B_n(s) = \left( \frac{s_B}{s_B + s} \right)^n .

    This creates a space-like pole (of degree `n`) at :math:`s = -s_B`.
    """
    return (sB / (sB + s)) ** n


def BreitWigner(s, mass: float, width: float):
    r"""
    Breit–Wigner distribution.

    Parameters
    ----------
    s : array_like
        Four-momentum squared of the propagating particle.
    mass : float
        Mass of the propagating particle.
    width : float
        Width of the propagating particle.

    Returns
    -------
    G : array_like
        The same shape as input `s`.

    Notes
    -----
    The distribution is defined as :cite:`Breit:1936zzb`

    .. math:: G(s, M, \Gamma) = \frac{1}{s - M^2 + i M \Gamma},

    where :math:`M` and :math:`\Gamma` are the mass and the
    width of the propagating particle, correspondingly.
    """
    return 1 / (s - mass**2 + 1j * mass * width)


def BreitWignerED(s, mass: float, width_s: callable, width_0: float):
    r"""
    Breit–Wigner distribution.

    Parameters
    ----------
    s : array_like
        Four-momentum squared of the propagating particle.
    mass : float
        Mass of the propagating particle.
    width_s : callable
        Energy-dependent width of the propagating particle, must be a function of `s`, `mass`, and `width_0`.
    width_0 : float
        A parameter that enters the definition of the energy-dependent width.

    Returns
    -------
    G : array_like
        The same shape as input `s`.

    Notes
    -----
    The distribution is defined as

    .. math:: G(s, M, \Gamma_s, \Gamma_0) = \frac{1}{s - M^2 + i M \Gamma_s(s, M, \Gamma_0)},

    where :math:`M` and :math:`\Gamma_s` are the mass and the
    width of the propagating particle, correspondingly. The index :math:`s` indicates that
    the width need not be constant and can depend on energy.

    See Also
    --------
    dispersionrelations.kinematics.BreitWigner
    """
    return 1 / (s - mass**2 + 1j * mass * width_s(s, mass, width_0))


def radiative_width_to_normalisation_squared(width, mV, mP):
    r"""Computes the vector–pseudoscalar form factor normalisation squared from the radiative width of the vector particle.

    Parameters
    ----------
    width : float
        Radiative width of the vector particle.
    mV : float
        Mass of the vector particle.
    mP : float
        Mass of the pseudoscalar particle.

    Returns
    -------
    c_squared : float
        Normalisation squared :math:`|f(0)|^2`.

    Notes
    -----

    The radiative width :math:`\Gamma` is connected to the form factor normalisation :math:`|f(0)|` via

    .. math:: \Gamma_{V \rightarrow P\gamma} = \frac{\alpha(M_V^2 - M_P^2)^3}{24 M_V^3} |f_{VP}(0)|^2.
    
    """
    return width / ((alpha_fs / 24) * ((mV**2 - mP**2) / mV) ** 3)


def radiative_width_to_normalisation_squared_error(
    width, mV, mP, width_std, mV_std, mP_std
):
    r"""Computes the error of the vector–pseudoscalar form factor normalisation squared from the radiative width of the vector particle.

    Parameters
    ----------
    width, width_std : float
        Radiative width of the vector particle (and the corresponding error).
    mV, mV_std : float
        Mass of the vector particle (and the corresponding error).
    mP, mP_std : float
        Mass of the pseudoscalar particle (and the corresponding error).
    Returns
    -------
    c_squared_std : float
        Error of the normalisation squared :math:`|f(0)|^2`.

    Notes
    -----

    The radiative width :math:`\Gamma` is connected to the form factor normalisation :math:`|f(0)|` via

    .. math:: \Gamma_{V \rightarrow P\gamma} = \frac{\alpha(M_V^2 - M_P^2)^3}{24 M_V^3} |f_{VP}(0)|^2.

    """
    jacobian_width = 1 / ((alpha_fs / 24) * ((mV**2 - mP**2) / mV) ** 3)
    jacobian_mV = (
        3 * width * mV**2 * (mP**2 + mV**2) / (alpha_fs * (mV**2 - mP**2) ** 4)
    )
    jacobian_mP = 6 * width * mP * mV**3 / (alpha_fs * (mV**2 - mP**2) ** 4)
    return np.sqrt(
        (jacobian_width * width_std) ** 2
        + (jacobian_mV * mV_std) ** 2
        + (jacobian_mP * mP_std) ** 2
    )


class Channel(DispersionIntegralRHC, ABC):
    r"""
    Dispersive representation of a channel.

    Parameters
    ----------
    loop_integrand : callable
        Imaginary part of the self-energy.
    threshold : float
        Lower boundary of the integral.
    infinity : float, optional
        Upper boundary of the integral in the units of :code:`threshold`.
    integration_split_points : array_like, optional
        Integration split points in the units of :code:`threshold`.
    integration_order : int, optional
        Gauss–Legendre quadrature order.
    subtraction_point : float, optional
        Point :math:`s_0` of subtraction.
    subtraction_constants : array_like, optional
        Array of :math:`f_i` subtraction constants. Defines subtraction level :math:`n`.

    Notes
    -----
    The integral is defined as

    .. math:: \Pi(s) = \sum_{i=0}^{n-1}f_i (s-s_0)^i + \frac{(s-s_0)^n}{\pi}\int_{s_{thr}}^{\infty}\frac{\mathrm{Im}(\Pi(x)) \, dx}{(x-s_0)^n(x-s-i\epsilon)},

    where :math:`s_0` is the subtraction point, :math:`f_i` are the subtraction constants,
    :math:`n` is the subtraction level, and
    :math:`\mathrm{Im}(\Pi(x))` is the `integrand`.

    See also
    --------

    dispersionrelations.dynamics.StableTwoBodyChannel

    """

    def __init__(
        self,
        loop_integrand: callable,
        threshold: float,
        infinity: float = 1e6,
        integration_split_points: list[float] = [
            2,
            3,
            5,
            8,
            10,
            50,
            1e2,
            1e3,
            1e4,
            1e5,
        ],
        integration_order: int = 100,
        subtraction_point: float = 0,
        subtraction_constants: list[float] = [0],
    ):
        super().__init__(
            loop_integrand,
            threshold,
            infinity,
            integration_split_points,
            integration_order,
            subtraction_point,
            subtraction_constants,
        )

    def loop_discontinuity(self, s):
        r"""
        Discontinuity of the two-body channel self-energy.

        Parameters
        ----------
        s : array_like
            Four-momentum squared of the two-body state.

        Returns
        -------
        DiscΠ : array_like
            The same shape as input `s`.

        Notes
        -----
        
        We assume the Schwarz reflection principle and therefore,

        .. math::
            \mathrm{Disc}(\Pi(s)) = 2i \, \mathrm{Im}(\Pi(s)) \, .

        """
        result_up = np.where(np.imag(s) >= 0, 2j * self.integrand(s), 0)
        result_down = np.where(
            np.imag(s) < 0, np.conj(2j * self.integrand(np.conj(s))), 0
        )
        return result_up + result_down

    def loop(self, s, sheet=1):
        r"""
        Invokes the :code:`__call__` method of the parent class,
        with an added functionality of switching to the second sheet.
        Computed the self-energy function of the two-body channel.

        Parameters
        ----------
        s : array_like
            Four-momentum squared of the two-body state.
        sheet : int, optional
            Riemann sheet number (can only be 1 or 2, other inputs return the first sheet).

        Returns
        -------
        Π : array_like
            The same shape as input `s`.
        """
        result = self.__call__(s)
        if sheet == 2:
            result -= self.loop_discontinuity(s)
        return result

    @abstractmethod
    def phase_space_function(self, s):
        return None

    @abstractmethod
    def vertex_function_squared(self, s):
        return None

    @abstractmethod
    def taming_factor_squared(self, s):
        return None

    def vertex_function(self, s):
        return np.sqrt(self.vertex_function_squared(s) + 0j)

    def taming_factor(self, s):
        return np.sqrt(self.taming_factor_squared(s) + 0j)


class PrecomputedChannel(Channel):
    def __init__(self, s_values, integrand_values, threshold, **kwargs):
        self.last_s = s_values[-1]
        self.last_integrand = integrand_values[-1]
        self.interpolated_integrand = interp1d(s_values, integrand_values)
        warnings.warn(
            "You are loading this channel from precomputed values. Full discontinuity will be absorbed in the phase space function."
        )
        super().__init__(self.extrapolated_integrand, threshold, **kwargs)

    def extrapolated_integrand(self, s):
        s = np.asarray(s)
        res = np.zeros_like(s)
        mask1 = s < self.threshold
        mask2 = s > self.last_s
        mask3 = ~(mask1 | mask2)
        res[mask1] = 0
        res[mask2] = self.last_integrand
        res[mask3] = self.interpolated_integrand(s[mask3])
        return res

    def phase_space_function(self, s):
        return self.integrand(s)

    def vertex_function_squared(self, s):
        return 0 * s + 1

    def taming_factor_squared(self, s):
        return 0 * s + 1


class StableTwoBodyChannel(Channel):
    r"""
    Constructs a two-body channel with stable particles with masses :code:`m1` and :code:`m2`.

    Parameters
    ----------
    m1 : float
        Mass of the first particle.
    m2 : float
        Mass of the second particle.
    vertex_function_squared : callable
        Vertex function squared. Depends on the interaction of the particles.
    taming_factor : callable
        Taming factor, to make sure that the integrand does not grow indefinitely.
    **kwargs
        Other keyword arguments, to be passed to the constructor of the parent class.

    Notes
    -----

    The integrand of the loop is constructed via

    .. math::
        \mathrm{Disc}(\Pi(s)) = 2i \, \rho(s, m_1^2, m_2^2) \beta^2(s, m_1^2, m_2^2) B^2(s) \, ,

    where :math:`\rho` is the phase space, :math:`\beta` is the vertex function,
    and :math:`B` is the taming factor.
    """

    def __init__(
        self,
        m1,
        m2,
        vertex_function_squared=vertex_VPP__2,
        taming_factor_squared=lambda s: taming_spacelike_pole(s) ** 2,
        **kwargs,
    ):
        # Store parameters:
        self.m1 = m1
        self.m2 = m2
        self.threshold = (m1 + m2) ** 2
        # Implement abstract methods:
        self.__psf__ = np.vectorize(lambda s: phase_space_twobody(s, m1, m2))
        self.__vfs__ = np.vectorize(lambda s: vertex_function_squared(s, m1, m2))
        self.__tfs__ = np.vectorize(lambda s: taming_factor_squared(s))
        # Construct the integrand and call the parent constructor:
        loop_integrand = (
            lambda s: self.phase_space_function(s)
            * self.vertex_function_squared(s)
            * self.taming_factor_squared(s)
        )
        super().__init__(loop_integrand, self.threshold, **kwargs)

    def phase_space_function(self, s):
        return self.__psf__(s)

    def vertex_function_squared(self, s):
        return self.__vfs__(s)

    def taming_factor_squared(self, s):
        return self.__tfs__(s)


class UnstableParticle:
    r"""
    Unstable particle.

    Parameters
    ----------
    decay_channel : TwoBodyChannel
        A channel that the particle decays into.
        The channel does not need to be two-body,
        as long as it has a :code:`threshold` and
        functions :code:`integrand` and :code:`loop`.
    pole_location : complex
        Location of the resonance pole (see :func:`sR`).
    pole_sheet : int
        Riemann sheet of the decay channel, on which the resonance lives.
        For above-threshold resonances, the typical sheet is `2`.
    mass_and_coupling : (float, float)
        Bare mass and the coupling constant, parameters of the propagator.
        If passed, :code:`pole_location` is ignored.
    interpolation_points : list
        Intervals to use for interpolating the spectral function.
    interpolation_density : int
        Number of points per interpolation interval.
    """

    def __init__(
        self,
        decay_channel: Channel,
        pole_location: complex,
        pole_sheet: int = 2,
        mass_and_coupling: Optional[tuple] = None,
        interpolation_points=[1, 2, 5, 1e1, 5e1, 1e2, 1e3, 1e5, 1e8],
        interpolation_density=1000,
    ):
        self.decay_channel = decay_channel
        self.sp = pole_location
        self.pole_sheet = pole_sheet
        self.mass_and_coupling = mass_and_coupling
        self.__calculate_m_g__()
        self.interpolation_points = interpolation_points
        self.interpolation_density = interpolation_density
        self.__interpolate_spectral_function__()

    def __calculate_m_g__(self):
        if self.mass_and_coupling is not None:
            self.m, self.g = self.mass_and_coupling
            return

        # The code below is the only part where self.decay_channel.loop should be known for complex numbers.
        # If mass_and_coupling is provided, real-valued loop is sufficient.
        loop_at_the_pole = self.decay_channel.loop(self.sp, sheet=self.pole_sheet)
        self.g = np.sqrt(-np.imag(self.sp) / np.imag(loop_at_the_pole))
        self.m = np.sqrt(np.real(self.sp) + self.g**2 * np.real(loop_at_the_pole))

    def propagator(self, s, sheet=1):
        r"""
        Propagator of the unstable particle.

        Parameters
        ----------
        
        s : array_like
            Four-momentum squared of the propagating particle.
        sheet : int, optional
            Riemann sheet number of the decay channel self-energy
            (can only be 1 or 2, other inputs return the first sheet).

        Returns
        -------
        G : array_like
            The same shape as input `s`.

        Notes
        -----

        The propagator is defined as

        .. math:: G(s) = \frac{1}{s - m^2 + g^2 \Pi(s)} \, ,

        where :math:`\Pi(s)` is the self-energy of the decay channel.
        """
        return 1 / (s - self.m**2 + self.g**2 * self.decay_channel.loop(s, sheet))

    def spectral_function_CP(self, s):
        r"""
        Spectral function of the unstable particle for :math:`s\in\mathbb{C}`.

        Parameters
        ----------
        
        s : array_like
            Four-momentum squared of the propagating particle.

        Returns
        -------
        σ : array_like
            The same shape as input `s`.

        Notes
        -----

        The spectral function is defined as

        .. math:: \sigma(s) = \frac{1}{\pi} G^{I}(s) G^{II}(s) \, g^2 \mathrm{Im}(\Pi(s)) \, ,

        which, for real :math:`s` reduces to
        
        .. math:: \sigma(s\in\mathbb{R}) = -\frac{1}{\pi} \mathrm{Im}(G(s)) \, .
        """
        return (
            1
            / np.pi
            * self.propagator(s, sheet=1)
            * self.propagator(s, sheet=2)
            * self.g**2
            * self.decay_channel.integrand(s)
        )

    def __interpolate_spectral_function__(self):
        s_interpolation = np.concatenate(
            [
                self.decay_channel.threshold
                * np.linspace(
                    1.0001 * self.interpolation_points[i],
                    self.interpolation_points[i + 1],
                    self.interpolation_density,
                )
                for i in range(len(self.interpolation_points) - 1)
            ]
        )
        sf_interpolation = np.real(self.spectral_function_CP(s_interpolation))
        self.__spectral_function_interpolated__ = interp1d(
            [self.decay_channel.threshold, *s_interpolation], [0, *sf_interpolation]
        )

    def spectral_function_RE(self, s):
        r"""
        Spectral function of the unstable particle for :math:`s\in\mathbb{R}`.

        Parameters
        ----------
        
        s : array_like
            Four-momentum squared of the propagating particle.

        Returns
        -------
        σ : array_like
            The same shape as input `s`.

        Notes
        -----

        The spectral function is defined as

        .. math:: \sigma(s) = -\frac{1}{\pi} \mathrm{Im}(G(s)) \, .

        An interpolation function is used when possible.
        """
        below_threshold = np.real(s) < self.decay_channel.threshold
        above_interpolation_maximum = (
            np.real(s) > self.interpolation_points[-1] * self.decay_channel.threshold
        )
        interpolatable = ~below_threshold & ~above_interpolation_maximum
        result = np.zeros_like(s, dtype=float)
        if np.sum(interpolatable) > 0:
            result[interpolatable] = self.__spectral_function_interpolated__(
                s[interpolatable]
            )
        return result


class SemiStableTwoBodyCut:
    r"""
    Two-body channel with one unstable particle and one spectator.

    Parameters
    ----------
    unstable_particle : UnstableParticle
        Unstable particle.
    spectator_mass : float
        Mass of the spectator particle.
    vertex_function : callable
        Vertex function. Depends on the interaction of the particles.
    phase_space : callable
        Phase space function.
    taming_factor : callable
        Taming factor, to make sure that the integrand does not grow indefinitely.
    gl_order : int
        Gauss–Legendre quadrature order.
    split_n : int
        Number of intervals per numerical integration.
    interpolation_points : list
        Intervals to use for interpolating the spectral function.
    interpolation_density: int
        Number of points per interpolation interval.
    """

    def __init__(
        self,
        unstable_particle: UnstableParticle,
        spectator_mass: float,
        vertex_function_squared: callable,
        gl_order: int = 100,
        split_n: int = 3,
        integration_split_points: list[float] = [2, 5, 10, 50, 100],
        interpolation_points: list = [
            1,
            2,
            3,
            4,
            5,
            1e1,
            3e1,
            5e1,
            1e2,
            1e3,
            1e4,
            1e5,
            1e6,
            1e7,
            1e8,
            1e9,
        ],
        interpolation_density: int = 1000,
    ):
        self.unstable_particle = unstable_particle
        self.spectator_mass = spectator_mass
        self.phase_space_factor = phase_space_twobody
        self.vertex_function_squared = vertex_function_squared
        self.gl_order = gl_order
        self.split_n = split_n
        self.integration_split_points = (
            self.unstable_particle.decay_channel.threshold
            * np.array(integration_split_points)
        )
        self.interpolation_points = interpolation_points
        self.interpolation_density = interpolation_density
        self.__interpolate_ImPI_integral__()

    def ImPI_integrand_RE(self, s, x):
        r"""
        Integrand for the :math:`\mathrm{Im}(\Pi(s))` spectral integral for :math:`x\in\mathbb{R}`.

        Parameters
        ----------
        s : complex
            Four-momentum squared of the two-body system.
        x : array_like
            Four-momentum squared of the unstable particle.

        Returns
        -------
        f : array_like
            The same shape as input `x`.

        Notes
        -----
        The integrand is defined via

        .. math:: \mathrm{Im}(\Pi(s)) = \int_{s_\text{thr}}^{(\sqrt{s}-m)^2} \sigma(x) \rho(s, x, m^2) \beta^2(s, x, m^2) B^2(s) \, ,

        where :math:`\sigma` is the spectral function of the unstable particle
        and :math:`s_\text{thr}` is the threshold of the corresponding decay channel.
        """
        return (
            self.unstable_particle.spectral_function_RE(x)
            * self.phase_space_factor(s, np.sqrt(x + 0j), self.spectator_mass)
            * self.vertex_function_squared(s, np.sqrt(x + 0j), self.spectator_mass)
        )

    def ImPI_integrand_CP(self, s, x):
        r"""
        Integrand for the :math:`\mathrm{Im}(\Pi(s))` spectral integral for :math:`x\in\mathbb{C}`.

        Parameters
        ----------
        s : complex
            Four-momentum squared of the two-body system.
        x : array_like
            Four-momentum squared of the unstable particle.

        Returns
        -------
        f : array_like
            The same shape as input `x`.

        Notes
        -----
        The integrand is defined via

        .. math:: \mathrm{Im}(\Pi(s)) = \int_{s_\text{thr}}^{(\sqrt{s}-m)^2} \sigma(x) \rho(s, x, m^2) \beta^2(s, x, m^2) B^2(s) \, ,

        where :math:`\sigma` is the spectral function of the unstable particle
        and :math:`s_\text{thr}` is the threshold of the corresponding decay channel.
        """
        return (
            self.unstable_particle.spectral_function_CP(x)
            * self.phase_space_factor(s, np.sqrt(x + 0j), self.spectator_mass)
            * self.vertex_function_squared(s, np.sqrt(x + 0j), self.spectator_mass)
        )

    def ImPI_integral(self, s):
        r"""
        The :math:`\mathrm{Im}(\Pi(s))` spectral integral.

        Parameters
        ----------
        s : complex
            Four-momentum squared of the two-body system.

        Returns
        -------
        ImΠ : complex
            A single complex number.

        Notes
        -----
        The integral is defined via

        .. math:: \mathrm{Im}(\Pi(s)) = \int_{s_\text{thr}}^{(\sqrt{s}-m)^2} \sigma(x) \rho(s, x, m^2) \beta^2(s, x, m^2) B^2(s) \, ,

        where :math:`\sigma` is the spectral function of the unstable particle
        and :math:`s_\text{thr}` is the threshold of the corresponding decay channel.

        For complex values of :math:`s`, the integral contour is taken to be
        rectangular (see e.g. :cite:`JPAC:2018zwp` for more details).
        """
        lower_bound = self.unstable_particle.decay_channel.threshold
        upper_bound = (np.sqrt(s) - self.spectator_mass) ** 2
        middle_bound = np.real(upper_bound)

        integrand = lambda x: self.ImPI_integrand_RE(s, x)
        if np.imag(upper_bound) != 0:
            integrand = lambda x: self.ImPI_integrand_CP(s, x)

        integration_points_1 = np.concatenate(
            (
                [lower_bound],
                self.integration_split_points[
                    self.integration_split_points < middle_bound
                ],
                [middle_bound],
            )
        )

        integral_1 = 0 * s + 0j
        for i in range(len(integration_points_1) - 1):
            integral_1 += integrate_gl(
                integrand,
                integration_points_1[i],
                integration_points_1[i + 1],
                order=self.gl_order,
                split_n=self.split_n,
            )
        # integral_1 = integrate_gl(
        #     integrand,
        #     lower_bound,
        #     middle_bound,
        #     order=self.gl_order,
        #     split_n=self.split_n,
        # )

        if np.imag(upper_bound) == 0:
            return integral_1

        integral_2 = integrate_gl(
            integrand,
            lower_bound,
            middle_bound,
            order=self.gl_order,
            split_n=self.split_n,
        )

        return integral_1 + integral_2

    @property
    def ImPI_integral_vectorized(self):
        r"""Vectorized version of :func:`dispersionrelations.dynamics.SemiStableTwoBodyCut.ImPI_integral`"""
        return np.vectorize(self.ImPI_integral)

    def __interpolate_ImPI_integral__(self):
        s_interpolation = np.concatenate(
            [
                self.unstable_particle.decay_channel.threshold
                * np.linspace(
                    1.0001 * self.interpolation_points[i],
                    self.interpolation_points[i + 1],
                    self.interpolation_density,
                )
                for i in range(len(self.interpolation_points) - 1)
            ]
        )
        ImPI_interpolation = np.array(
            [np.real(self.ImPI_integral(s)) for s in s_interpolation]
        )
        self.__ImPI_integral_interpolated__ = interp1d(
            [self.unstable_particle.decay_channel.threshold, *s_interpolation],
            [0, *ImPI_interpolation],
        )

    def ImPI_integral_RE(self, s):
        r"""
        The :math:`\mathrm{Im}(\Pi(s))` spectral integral for :math:`s\in\mathbb{R}`.

        Parameters
        ----------
        s : array_like
            Four-momentum squared of the two-body system.

        Returns
        -------
        ImΠ : array_like
            The same shape as input `s`.

        Notes
        -----
        An interpolation function is used when possible.
        """
        below_threshold = np.real(s) < self.unstable_particle.decay_channel.threshold
        above_interpolation_maximum = (
            np.real(s)
            > self.interpolation_points[-1]
            * self.unstable_particle.decay_channel.threshold
        )
        interpolatable = ~below_threshold & ~above_interpolation_maximum
        result = np.zeros_like(s, dtype=float)
        if np.sum(interpolatable) > 0:
            result[interpolatable] = self.__ImPI_integral_interpolated__(
                s[interpolatable]
            )
        if np.sum(above_interpolation_maximum) > 0:
            result[above_interpolation_maximum] = self.ImPI_integral_vectorized(
                s[above_interpolation_maximum]
            )
        return result

    def ImPI_integral_CP(self, s):
        r"""
        The :math:`\mathrm{Im}(\Pi(s))` spectral integral for :math:`s\in\mathbb{C}`.

        Parameters
        ----------
        s : array_like
            Four-momentum squared of the two-body system.

        Returns
        -------
        ImΠ : array_like
            The same shape as input `s`.
        """
        return self.ImPI_integral_vectorized(s)

    def __call__(self, s):
        r"""
        The :math:`\mathrm{Im}(\Pi(s))` spectral integral for a generic :math:`s`.

        Parameters
        ----------
        s : array_like
            Four-momentum squared of the two-body system.

        Returns
        -------
        ImΠ : array_like
            The same shape as input `s`.

        Notes
        -----
        An interpolation function is used when possible for real values of :math:`s`.
        """
        s = np.array(s)
        is_complex = np.imag(s) != 0
        result = np.zeros(s.shape, dtype=np.complex128)
        if np.sum(is_complex) > 0:
            result[is_complex] = self.ImPI_integral_CP(s[is_complex])
        if np.sum(~is_complex) > 0:
            result[~is_complex] = self.ImPI_integral_RE(np.real(s[~is_complex]))
        return result


class SemiStableTwoBodyChannel(Channel):
    def __init__(
        self,
        twobodycut: SemiStableTwoBodyCut,
        taming_factor_squared=lambda s: taming_spacelike_pole(s) ** 2,
    ):
        # Store parameters:
        threshold = (
            np.sqrt(twobodycut.unstable_particle.decay_channel.threshold)
            + twobodycut.spectator_mass
        ) ** 2
        # Implement abstract methods:
        self.__psf__ = twobodycut.__call__
        self.__vfs__ = np.vectorize(lambda s: 1)
        self.__tfs__ = np.vectorize(taming_factor_squared)
        # Construct the integrand and call the parent constructor:
        loop_integrand = (
            lambda s: self.phase_space_function(s)
            * self.vertex_function_squared(s)
            * self.taming_factor_squared(s)
        )
        super().__init__(loop_integrand, threshold)

    def phase_space_function(self, s):
        return self.__psf__(s)

    def vertex_function_squared(self, s):
        return self.__vfs__(s)

    def taming_factor_squared(self, s):
        return self.__tfs__(s)


class TwoPotentialModel:
    def __init__(
        self,
        channels: dict[str, Channel],
        bg_par: dict,
        res_par: dict,
        channel_sheets: dict = {},
        s_array: Union[np.ndarray, list] = np.linspace(E_almost_zero, 3 * GeV, 200)
        ** 2,
    ):
        # Initialize channels
        self.channels = channels
        self.channel_sheets = {key: 1 for key in channels.keys()}
        for key in channel_sheets.keys():
            if key in channels.keys():
                self.channel_sheets[key] = channel_sheets[key]
            else:
                warnings.warn(
                    f"Channel '{key}' not found in the provided channels dictionary. Ignoring the sheet specification."
                )
        # Initialize parameters
        self.bg_par = bg_par
        self.res_par = res_par
        # Initialize auxilary variables
        self.nC = len(channels)
        self.nR = len(res_par["m"])
        self.IdC = np.identity(self.nC)
        self.IdR = np.identity(self.nR)

        # Initialize kinematic variables
        self.__set_s_array__(s_array)

        # Default parameters, in case none are provided
        self.bg_par = {
            "f0": np.zeros((self.nC, self.nC)),
            "fR": np.array([0.0]),
        }
        self.res_par = {
            "g": np.zeros((self.nR, self.nC)),
            "m": np.arange(self.nR) + 1 * GeV,
            "c": np.ones(self.nC),
            "a": np.ones(self.nR),
        }

        # Store model parameters from passed dictionaries
        self.store_parameters(bg_par, res_par)

        # Computation steps
        self.precompute()
        self.compute_bg()
        self.compute_res()

    def __set_s_array__(self, s_array):
        # Index _rs stands for "reshaped"
        self.s_array = np.array(s_array)
        self.s_rs = self.s_array[..., np.newaxis, np.newaxis]
        self.E_array = np.sign(s_array) * np.sqrt(np.absolute(self.s_array))
        self.E_rs = self.E_array[..., np.newaxis, np.newaxis]

    def store_parameters(self, new_bg_par=None, new_res_par=None):
        if new_bg_par is not None:
            for key in new_bg_par:
                if key in self.bg_par.keys():
                    self.bg_par[key] = new_bg_par[key]
                else:
                    warnings.warn(
                        f"Background parameter '{key}' not recognized. Ignoring."
                    )

            self.f0 = np.array(self.bg_par["f0"])
            self.fR = np.array(self.bg_par["fR"])

        if new_res_par is not None:
            for key in new_res_par:
                if key in self.res_par.keys():
                    self.res_par[key] = new_res_par[key]
                else:
                    warnings.warn(
                        f"Resonance parameter '{key}' not recognized. Ignoring."
                    )

            self.g = np.array(self.res_par["g"])
            self.m = np.array(self.res_par["m"])
            self.c = np.array(self.res_par["c"])
            self.a = np.array(self.res_par["a"])

    def __diag_in_channel_space__(self, arr):
        arr = np.asarray(arr)
        if arr.shape != (*self.s_array.shape, self.nC):
            raise ValueError(
                f"Input array must have shape {(*self.s_array.shape, self.nC)}"
            )

        channel_indexing = np.arange(self.nC)
        result = np.zeros((*self.s_array.shape, self.nC, self.nC)) + 0j
        result[..., channel_indexing, channel_indexing] = arr
        return result

    def __diag_in_resonance_space__(self, arr):
        arr = np.asarray(arr)
        if arr.shape != (*self.s_array.shape, self.nR):
            raise ValueError(
                f"Input array must have shape {(*self.s_array.shape, self.nR)}"
            )

        resonance_indexing = np.arange(self.nR)
        result = np.zeros((*self.s_array.shape, self.nR, self.nR)) + 0j
        result[..., resonance_indexing, resonance_indexing] = arr
        return result

    def precompute(self):
        loops_arr = np.array(
            [
                self.channels[key].loop(self.s_array, sheet=self.channel_sheets[key])
                for key in self.channels.keys()
            ]
        )
        self.loops = self.__diag_in_channel_space__(np.moveaxis(loops_arr, 0, -1))

    def compute_bg(self, new_bg_par=None):
        if new_bg_par is not None:
            self.store_parameters(new_bg_par=new_bg_par)

        self.f_s = (
            np.ones((*self.s_array.shape, self.nC, self.nC)) + self.fR * self.s_rs
        )
        self.VB = self.f0 / self.f_s
        self.TB = np.linalg.solve(self.IdC - self.VB @ self.loops, self.VB)
        self.vertex = self.IdC + self.TB @ self.loops
        self.loops_with_bg = self.loops @ self.vertex

    def compute_res(self, new_res_par=None):
        if new_res_par is not None:
            self.store_parameters(new_res_par=new_res_par)

        self.GR_inv = self.__diag_in_resonance_space__(
            self.s_array[..., np.newaxis] - self.m[np.newaxis, ...] ** 2
        )
        self.GR_renormalized_inv = self.GR_inv + self.g @ self.loops_with_bg @ self.g.T
        self.tR = -self.g.T @ np.linalg.solve(self.GR_renormalized_inv, self.g) # TODO: add a warning for failed inversion
        self.TR = self.vertex @ self.tR @ (np.swapaxes(self.vertex, -2, -1))
        self.F = np.einsum(
            "...ij,...j->...i",
            -self.vertex @ self.g.T,
            np.linalg.solve(self.GR_renormalized_inv, self.a),  # TODO: add a warning for failed inversion
        )

    def compute_everything(self, bg_par=None, res_par=None):
        self.compute_bg(bg_par)
        self.compute_res(res_par)

    def new_s_array(self, s_array):
        self.__set_s_array__(s_array)
        self.precompute()
        self.compute_bg()
        self.compute_res()

    def __amplitude__(self, amp):
        vertex_factors = self.__diag_in_channel_space__(
            np.moveaxis(
                np.array(
                    [
                        channel.vertex_function(self.s_array)
                        * channel.taming_factor(self.s_array)
                        for channel in self.channels.values()
                    ]
                ),
                0,
                -1,
            )
        )
        return vertex_factors @ amp @ vertex_factors

    def T_matrix_bg(self):
        return self.__amplitude__(self.TB)

    def T_matrix_res(self):
        return self.__amplitude__(self.TR)

    def T_matrix(self):
        return self.__amplitude__(self.TB + self.TR)

    def S_matrix(self):
        phase_space_factors = self.__diag_in_channel_space__(
            np.moveaxis(
                np.array(
                    [
                        channel.phase_space_function(self.s_array)
                        for channel in self.channels.values()
                    ]
                ),
                0,
                -1,
            )
        )
        return self.IdC + 2j * phase_space_factors * self.T_matrix()

    def form_factor(self):
        return self.F


import numpy as np
from skimage import measure

import matplotlib.pyplot as plt
from matplotlib import colors
from matplotlib.pyplot import Axes
from mpl_toolkits.mplot3d.axes3d import Axes3D


plt.rcParams["figure.figsize"] = [8, 4]
plt.rcParams["figure.dpi"] = 100

plt.rc("font", **{"family": "serif", "serif": ["Computer Modern"]})
plt.rc("text", usetex=True)


def prepare_re_im(
    re_min,
    re_max,
    re_n,
    im_min,
    im_max,
    im_n,
    im_near_zero=1e-2, # relative units
    include_im_zero=True,
):
    z_re = np.linspace(re_min, re_max, re_n)
    z_im = np.linspace(im_min, im_max, im_n)
    if im_near_zero != 0.0:
        z_im = np.concat(
            (
                np.linspace(im_min, -np.abs(im_near_zero * im_max), im_n // 2),
                np.linspace(0, 0, int(include_im_zero)),
                np.linspace(np.abs(im_near_zero * im_max), im_max, im_n // 2),
            )
        )
    return z_re, z_im


def prepare_complex_plot_input(function, z_re, z_im, label):
    plot_input = {
        "function": function,
        "label": label,
        "z_re": z_re,
        "z_im": z_im,
    }

    plot_input["z_RE"], plot_input["z_IM"] = np.meshgrid(z_re, z_im)
    plot_input["z_CP"] = plot_input["z_RE"] + 1j * plot_input["z_IM"]
    plot_input["q_CP"] = function(plot_input["z_CP"])

    return plot_input


def figure_3x2(**figkwargs):
    fig = plt.figure(**figkwargs)

    ax_re = fig.add_subplot(2, 3, 1)
    ax_im = fig.add_subplot(2, 3, 2)
    ax_abs = fig.add_subplot(2, 3, 3)
    ax_re_3d = fig.add_subplot(2, 3, 4, projection="3d", computed_zorder=False)
    ax_im_3d = fig.add_subplot(2, 3, 5, projection="3d", computed_zorder=False)
    ax_abs_3d = fig.add_subplot(2, 3, 6, projection="3d", computed_zorder=False)

    return fig


def plot_complex_ax(ax, plot_input, q_function, plotting_function = Axes.contourf, **pfkwargs):
    return plotting_function(ax, plot_input["z_RE"], plot_input["z_IM"], q_function(plot_input["q_CP"]), **pfkwargs)


def plot_contourf_ax(ax, plot_input, q_function, ncontours=20, centered_around_zero=False, **pfkwargs):
    q_current = q_function(plot_input['q_CP'])
    q_max = np.max(q_current)
    q_min = np.min(q_current)
    if centered_around_zero:
        q_max = max(np.abs(q_max), np.abs(q_min))
        q_min = -q_max
    contour_levels = np.linspace(q_min, q_max, ncontours)
    cf = plot_complex_ax(ax, plot_input, q_function, Axes.contourf, **dict(levels=contour_levels, **pfkwargs))
    plt.gcf().colorbar(cf, ax=ax)


def plot_curve_ax2d_ax3d(ax2d, ax3d, x, y, z, **pfkwargs):
    if ax2d is not None:
        ax2d.plot(x, y, **pfkwargs)

    if ax3d is not None:
        ax3d.plot(x, y, z, **pfkwargs)


def scatter_ax2d_ax3d(ax2d, ax3d, x, y, z, **pfkwargs):
    if ax2d is not None:
        ax2d.scatter(x, y, **pfkwargs)
        
    if ax3d is not None:
        ax3d.scatter(x, y, z, **pfkwargs)


def plot_ray_ax2d_ax3d(ax2d, ax3d, x, y, z, **pfkwargs):
    plot_curve_ax2d_ax3d(ax2d, ax3d, x, y, z, **pfkwargs)
    scatter_ax2d_ax3d(ax2d, ax3d, x[:1], y[:1], z[:1], **pfkwargs)


def plot_contours_ax2d_ax3d(ax2d, ax3d, plot_input, q_function, ncontours=20, centered_around_zero=False, **pfkwargs):
    q_current = q_function(plot_input['q_CP'])
    q_max = np.max(q_current)
    q_min = np.min(q_current)
    if centered_around_zero:
        q_max = max(np.abs(q_max), np.abs(q_min))
        q_min = -q_max
    for isovalue in np.linspace(q_min, q_max, ncontours):
        contours = measure.find_contours(q_function(plot_input["q_CP"]), isovalue)
        for contour in contours:
            x_cont = plot_input['z_re'][0] + (plot_input['z_re'][-1] - plot_input['z_re'][0]) * contour[:,1] / (len(plot_input['z_re']) - 1)
            y_cont = plot_input['z_im'][0] + (plot_input['z_im'][-1] - plot_input['z_im'][0]) * contour[:,0] / (len(plot_input['z_im']) - 1)
            z_cont = q_function(plot_input['function'](x_cont + 1j * y_cont))
            plot_curve_ax2d_ax3d(ax2d, ax3d, x_cont, y_cont, z_cont, **pfkwargs)


per_column_default = {
    0: {
        "q_function": np.real,
        "title": "Real part",
        "ncontours": 20,
        "centered_around_zero": True,
        "pfkwargs": dict(cmap="RdYlGn", norm=colors.CenteredNorm()),
    },
    1: {
        "q_function": np.imag,
        "title": "Imaginary part",
        "ncontours": 20,
        "centered_around_zero": True,
        "pfkwargs": dict(cmap="RdYlGn", norm=colors.CenteredNorm()),
    },
    2: {
        "q_function": np.abs,
        "title": "Absolute value",
        "ncontours": 24,
        "centered_around_zero": True,
        "pfkwargs": dict(cmap="Blues", norm=colors.Normalize(vmin=0)),
    }
}

def complex_plot_contours(plot_input, per_column=None):
    fig = figure_3x2(figsize=(10, 6))

    if per_column is None:
        per_column = per_column_default

    for col in per_column.keys():
        pc = per_column[col]
        fig.axes[col].set_title(pc['title'])
        plot_contourf_ax(fig.axes[col], plot_input, pc["q_function"], ncontours=pc["ncontours"], centered_around_zero=pc["centered_around_zero"], **pc["pfkwargs"])
        plot_complex_ax(fig.axes[3 + col], plot_input, pc["q_function"], Axes3D.plot_surface, **pc["pfkwargs"])
        plot_contours_ax2d_ax3d(fig.axes[col], fig.axes[3 + col], plot_input, pc["q_function"], ncontours=pc["ncontours"], centered_around_zero=pc["centered_around_zero"], c='white', alpha=0.3, lw=1)

    for ax in fig.axes[:3]:
        ax.axhline(0, c='white', alpha=0.1)
        ax.axvline(0, c='white', alpha=0.1)
        ax.set_aspect(1)

    fig.tight_layout()

    return fig


import pytest
import numpy as np
from dispersionrelations.utils import *

# # # # # # # # # # # # # # # # # # # # #
# HELPERS # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def assert_reflection(f, z, cut_angle=0.0, cut_in_imag=True):
    """Schwarz reflection principle test for functions with branch cuts."""
    z_refl = np.conj(z)
    rotation = np.exp(1j * cut_angle)
    z_rot = rotation * z
    z_refl_rot = rotation * z_refl
    f_val = f(z_rot)
    f_val_refl = f(z_refl_rot)
    if cut_in_imag:
        np.testing.assert_allclose(f_val, np.conj(f_val_refl))
    else:
        np.testing.assert_allclose(f_val, -np.conj(f_val_refl))


# # # # # # # # # # # # # # # # # # # # #
# TESTS # # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

z_positive_real_values = np.linspace(0, 1e4, 24)
z_negative_real_values = np.linspace(-1e4, 0, 24)
z_real_values = np.linspace(-1e4, 1e4, 42)
z_real_values_plus_imag = z_real_values + 1j * 1e-5
z_real_values_minus_imag = np.conj(z_real_values_plus_imag)


# # # # # # # # # # # # # # # # # # # # #
# sqrt_custom_branch_cut  # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def test_reflection_sqrt_custom_branch_cut():
    sample_points = z_real_values_plus_imag
    # np.random.seed(42)
    test_angle = (np.random.rand() - 0.5) * 2 * np.pi
    assert_reflection(
        lambda z: sqrt_custom_branch_cut(z, test_angle),
        sample_points,
        test_angle,
        cut_in_imag=False,
    )


# # # # # # # # # # # # # # # # # # # # #
# sqrtRHC # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def test_real_positive_values_sqrtRHC():
    sample_points = z_positive_real_values
    np.testing.assert_allclose(
        sqrtRHC(sample_points),
        np.sqrt(sample_points),
    )


def test_real_positive_values_sqrtRHC_sheet_2():
    sample_points = z_positive_real_values
    np.testing.assert_allclose(
        sqrtRHC(sample_points, sheet=2),
        -np.sqrt(sample_points),
    )


def test_real_negative_values_sqrtRHC():
    sample_points = z_negative_real_values
    np.testing.assert_allclose(
        sqrtRHC(sample_points),
        1j * np.sqrt(-sample_points + 0j),
    )


def test_real_negative_values_sqrtRHC_sheet_2():
    sample_points = z_negative_real_values
    np.testing.assert_allclose(
        sqrtRHC(sample_points, sheet=2),
        -1j * np.sqrt(-sample_points + 0j),
    )


def test_invalid_sheets_sqrtRHC():
    with pytest.raises(ValueError):
        invalid_sheets = [-1, 0, 3, None, "first", "second"]
        for sheet in invalid_sheets:
            _ = sqrtRHC(1.0, sheet=sheet)


def test_reflection_sqrtRHC():
    sample_points = z_real_values_plus_imag
    assert_reflection(sqrtRHC, sample_points, cut_in_imag=False)


# # # # # # # # # # # # # # # # # # # # #
# sqrtLHC # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def test_real_positive_values_sqrtLHC():
    sample_points = z_positive_real_values
    np.testing.assert_allclose(
        np.imag(sqrtLHC(sample_points)),
        np.sqrt(sample_points),
    )


def test_real_positive_values_sqrtLHC_sheet_2():
    sample_points = z_positive_real_values
    np.testing.assert_allclose(
        np.imag(sqrtLHC(sample_points, sheet=2)),
        -np.sqrt(sample_points),
    )


def test_real_negative_values_sqrtLHC():
    sample_points = z_negative_real_values
    np.testing.assert_allclose(
        sqrtLHC(sample_points),
        -1j * np.sqrt(sample_points + 0j),
    )


def test_real_negative_values_sqrtLHC_sheet_2():
    sample_points = z_negative_real_values
    np.testing.assert_allclose(
        sqrtLHC(sample_points, sheet=2),
        1j * np.sqrt(sample_points + 0j),
    )


def test_invalid_sheets_sqrtLHC():
    with pytest.raises(ValueError):
        invalid_sheets = [-1, 0, 3, None, "first", "second"]
        for sheet in invalid_sheets:
            _ = sqrtLHC(1.0, sheet=sheet)


def test_reflection_sqrtLHC():
    sample_points = z_real_values_plus_imag
    assert_reflection(sqrtLHC, sample_points, cut_in_imag=False)


# # # # # # # # # # # # # # # # # # # # #
# log_custom_branch_cut # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

def test_reflection_log_custom_branch_cut():
    sample_points = z_real_values_plus_imag
    # np.random.seed(42)
    test_angle = (np.random.rand() - 0.5) * 2 * np.pi
    assert_reflection(
        lambda z: log_custom_branch_cut(z, test_angle) - 1j * np.pi,
        sample_points,
        test_angle,
        cut_in_imag=True,
    )

# # # # # # # # # # # # # # # # # # # # #
# logRHC  # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

def test_real_positive_values_logRHC():
    sample_points = z_positive_real_values[z_positive_real_values != 0]
    np.testing.assert_allclose(
        logRHC(sample_points),
        np.log(sample_points),
    )

def test_real_negative_values_logRHC():
    sample_points = z_negative_real_values[z_negative_real_values != 0]
    np.testing.assert_allclose(
        logRHC(sample_points),
        np.log(-sample_points + 0j) + 1j * np.pi,
    )

def test_reflection_logRHC():
    sample_points = z_real_values_plus_imag
    assert_reflection(lambda z: logRHC(z) - 1j * np.pi, sample_points, cut_in_imag=True)

# # # # # # # # # # # # # # # # # # # # #
# logLHC  # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

def test_real_positive_values_logLHC():
    sample_points = z_positive_real_values[z_positive_real_values != 0]
    np.testing.assert_allclose(
        logLHC(sample_points),
        np.log(sample_points) + 1j * np.pi,
    )

def test_real_negative_values_logLHC():
    sample_points = z_negative_real_values[z_negative_real_values != 0]
    np.testing.assert_allclose(
        logLHC(sample_points),
        np.log(-sample_points + 0j),
    )

def test_reflection_logLHC():
    sample_points = z_real_values_plus_imag
    assert_reflection(lambda z: logLHC(z) - 1j * np.pi, sample_points, cut_in_imag=True)

# # # # # # # # # # # # # # # # # # # # #
# logC  # # # # # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

def test_real_positive_values_logC():
    sample_points = z_positive_real_values[z_positive_real_values != 0]
    np.testing.assert_allclose(
        logC(sample_points),
        np.log(sample_points),
    )

def test_real_negative_values_logC():
    sample_points = z_negative_real_values[z_negative_real_values != 0]
    np.testing.assert_allclose(
        logC(sample_points),
        np.log(-sample_points + 0j) - 1j * np.pi,
    )

def test_reflection_logC():
    sample_points = z_real_values_plus_imag
    assert_reflection(lambda z: logC(z) + 1j * np.pi, sample_points, cut_in_imag=True)


# # # # # # # # # # # # # # # # # # # # #
# extract_phase # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def test_extract_phase_positive_start():
    E_1 = np.linspace(0, 1.1, 1000)
    s_1 = E_1**2
    t_1 = 2 * np.pi * 3 * np.sin(2 * np.pi * s_1)
    t_2 = 0.0 * s_1
    t_3 = -t_1
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_1)), t_1)
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_2)), t_2)
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_3)), t_3)


def test_extract_phase_negative_start():
    E_1 = np.linspace(0, 1.1, 1000)
    s_1 = E_1**2
    t_1 = 2 * np.pi * 3 * np.sin(2 * np.pi * s_1)
    t_2 = 0.0 * s_1
    t_3 = -t_1
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_1)), t_1)
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_2)), t_2)
    np.testing.assert_allclose(extract_phase(np.exp(1j * t_3)), t_3)


# # # # # # # # # # # # # # # # # # # # #
# cite / uncite # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #

def test_cite_format():
    text = "reference"
    cited_text = cite(text)
    assert cited_text[:6] == r"\cite{" and cited_text[-1] == "}"


def test_cite_uncite():
    text = "reference"
    cited_text = cite(text)
    uncited_text = uncite(cited_text)
    assert text == uncited_text


def test_uncite_invalid_format():
    with pytest.raises(ValueError):
        invalid_citations = [
            r"text without \cite",
            r"\cite{missing_end",
            r"missing_start}",
            r"\\cite{accidental_double_backslash}",
        ]
        for invalid_citation in invalid_citations:
            _ = uncite(invalid_citation)


def test_empty_uncite():
    with pytest.warns(UserWarning):
        empty_citation = r"\cite{}"
        result = uncite(empty_citation)
        assert result == ""


# # # # # # # # # # # # # # # # # # # # #
# scientific_notation # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #
    

def test_scientific_notation():
    assert scientific_notation(np.pi) == '3.14'
    assert scientific_notation(np.pi, 10) == '3.1415926536'
    assert scientific_notation(13.4) == r'1.34 \times 10^{1}'
    assert scientific_notation(13.4, 10) == r'1.34 \times 10^{1}'
    assert scientific_notation(0.134) == r'1.34 \times 10^{-1}'
    assert scientific_notation(0.000134) == r'1.34 \times 10^{-4}'
    assert scientific_notation(13495) == r'1.35 \times 10^{4}'
    assert scientific_notation(13495, 20) == r'1.3495 \times 10^{4}'


# # # # # # # # # # # # # # # # # # # # #
# PDG_rounding  # # # # # # # # # # # # #
# # # # # # # # # # # # # # # # # # # # #


def test_PDG_rounding_case_1():
    a1, b1 = rounding_PDG(0.827, 0.367)
    assert a1 == 0.8 and b1 == 0.4

def test_PDG_rounding_case_1_integer():
    a1, b1 = rounding_PDG(827, 367)
    assert a1 == 800 and b1 == 400

def test_PDG_rounding_case_2():
    a2, b2 = rounding_PDG(0.827, 0.119)
    assert a2 == 0.83 and b2 == 0.12

def test_PDG_rounding_case_2_integer():
    a2, b2 = rounding_PDG(827, 119)
    assert a2 == 830 and b2 == 120

def test_PDG_rounding_case_3():
    a3, b3 = rounding_PDG(0.827, 0.971)
    assert a3 == 0.8 and b3 == 1.0

def test_PDG_rounding_case_3_integer():
    a3, b3 = rounding_PDG(827, 971)
    assert a3 == 800 and b3 == 1000


from dispersionrelations.dynamics import *

def test_thresholds():
    M1 = 0.5
    M2 = 0.2
    STHR = (M1 + M2)**2
    assert vertex_VPP(STHR, M1, M2) == 0.0
    assert vertex_VVP(STHR, M1, M2) == 0.0


def test_sR():
    M1 = 500
    G1 = 100
    s1 = sR(M1, G1)
    assert np.real(np.sqrt(s1)) == M1
    assert np.imag(np.sqrt(s1)) == - G1 / 2
    assert np.real(s1) == M1**2 - G1**2 / 4
    assert np.imag(s1) == - M1 * G1


import warnings

import numpy as np
from dispersionrelations.utils import logC

z_gl_100, w_gl_100 = np.polynomial.legendre.leggauss(100)
z_gl_500, w_gl_500 = np.polynomial.legendre.leggauss(500)
z_gl_1000, w_gl_1000 = np.polynomial.legendre.leggauss(1000)
z_gl_2000, w_gl_2000 = np.polynomial.legendre.leggauss(2000)

gl_storage = {
    100: (z_gl_100, w_gl_100),
    500: (z_gl_500, w_gl_500),
    1000: (z_gl_1000, w_gl_1000),
    2000: (z_gl_2000, w_gl_2000),
}


def integrate_gl(function : callable, a : float, b : float, order : int = 100, split_n : int = 3):
    r"""
    Gauss–Legendre integration.
    
    Parameters
    ----------
    function : callable
        A complex function of a single variable.
    a : float, complex
        Lower boundary of the integral.
    b : float, complex
        Upper boundary of the integral.
    order : int
        Quadrature order.
    split_n : int
        Number of subintervals of the same length :math:`[a,x], [x,y], ... [z,b]`.

    Returns
    -------
    I : float, complex
        Integral :math:`I = \int_a^b f(x) \mathrm{d}x`.
    """
    if order in gl_storage.keys():
        (gl_roots, gl_weights) = gl_storage[order]
    else:
        (gl_roots, gl_weights) = np.polynomial.legendre.leggauss(order)
    jacobian = (b - a) / 2 / split_n
    s_prime = np.array(
        [[(x + 1 + 2 * i) * jacobian + a for x in gl_roots] for i in range(split_n)]
    )
    return np.sum(function(s_prime) * gl_weights * jacobian)


def residue(function : callable, z0 : complex, radius : float = 1e-2, order : int = 100, split_n : int = 3):
    r"""
    Residue of a complex function.
    
    Parameters
    ----------
    function : callable
        Complex function of a single variable.
    z0 : complex
        Complex point to compute the residue at.
    radius : float
        Radius of the circular integral around :math:`z_0`.
    order : int
        Quadrature order.
    split_n : int
        Number of sections of the circle.

    Returns
    -------
    r : complex
        Residue, computed via :math:`r = \frac{1}{2\pi i}\oint_{\mathcal{C}_{z_0}} f(z) \mathrm{d}z`.

    Examples
    --------
    >>> from dispersionrelations import integrals
    >>> print(integrals.residue(lambda z: 13/(z-2-2j), z0=2+2j))
    np.complex128(13.000000000000007+4.3193350746692004e-15j)
    """
    function_θ = (
        lambda θ: 1
        / (2 * np.pi)
        * function(z0 + radius * np.exp(1j * θ))
        * radius
        * np.exp(1j * θ)
    )
    return integrate_gl(function_θ, 0, 2*np.pi, order, split_n)

class SimplePoleMonomial:
    r"""
    Cauchy integral of a shifted monomial

    Parameters
    ----------
    n : float
        Power of the shifted monomial (can be an integer or a half-integer).
        
    Raises
    ------
    NotImplementedError
        Any input :code:`n` other than an integer or a half-integer triggers an error.
    """
    def __init__(self, n : float):
        self.n = n
        if n % 1 == 0:
            if n == 0:
                self.indefinite_integral = self.__indefinite_integral_00__
            elif n > 0:
                self.indefinite_integral = self.__indefinite_integral_ZP__
            elif n < 0:
                self.indefinite_integral = self.__indefinite_integral_ZM__
            else:
                raise NotImplementedError(
                    f"Integral assignment failed for n={self.n}. Only integers and half-integers are supported."
                )
        elif n % 1 == 1 / 2:
            if n > 0:
                self.indefinite_integral = self.__indefinite_integral_ZP_HALF__
            elif n < 0:
                self.indefinite_integral = self.__indefinite_integral_ZM_HALF__
            else:
                raise NotImplementedError(
                    f"Integral assignment failed for n={self.n}. Only integers and half-integers are supported."
                )
        else:
            raise NotImplementedError(
                f"Integral assignment failed for n={self.n}. Only integers and half-integers are supported."
            )

    def __definite_integral__(self, s, s0, a, b):
        return self.indefinite_integral(s, s0, b) - self.indefinite_integral(s, s0, a)

    def __call__(self, s, s0, a, b):
        r"""
        Definite integral

        .. math::
            \mathcal{I}_{n}(s,s_0;a,b) = \int_a^b \frac{(x-s_0)^n \, \mathrm{d}x}{x-s}.

        Parameters
        ----------
        s : float
            Location of the Cauchy singularity.
        s0 : float
            Shift of the monomial in the integrand.
        a : float
            Lower integration boundary.
        b : float
            Upper integration boundary.

        Notes
        -----
        Several special cases are defined separately,

        Case 1. :math:`n = 0`,

        .. math:: \mathcal{I}_{0}(s, s_0; a, b) = \log(x-s) \bigg|_a^b \, .

        Case 2. :math:`n` is a positive integer,

        .. math::
            \mathcal{I}_{n \in \mathbb{Z}^+}(s, s_0; a, b) =
            (s-s_0)^n \left(
            \mathcal{I}_{0}(s, s_0; a, b)
            + \sum_{k=1}^{n} \frac{1}{k}\frac{(x-s_0)^{k}}{(s-s_0)^{k}}
            \bigg|_a^b
            \right) \, .

        Case 3. :math:`n = -\bar{n}` is a negative integer,

        .. math::
            \mathcal{I}_{-\bar{n}}(s, s_0; a, b)
            = (s-s_0)^{-\bar{n}} \left(
            \log\left(\frac{x-s}{x-s_0}\right)
            + \sum_{k=1}^{\bar{n}-1} \frac{1}{k} \frac{(s-s_0)^{k}}{(x-s_0)^{k}}
            \right)\bigg|_{a}^{b} \, .

        Case 4. :math:`n=\frac{m}{2}` is a positive half-integer,

        .. math::
            \mathcal{I}_{m/2}(s, s_0; a, b)
            = \sqrt{s-s_0}^{m} \left(
            \log\left(\frac{\sqrt{s-s_0}-\sqrt{x-s_0}}{\sqrt{s-s_0}+\sqrt{x-s_0}}\right)
            + \sum_{k=1}^{(m+1)/2} \frac{2}{2k-1} \frac{\sqrt{x-s_0}^{2k-1}}{\sqrt{s-s_0}^{2k-1}}
            \right)\bigg|_a^b \, .

        Case 5. :math:`n=-\frac{\bar{m}}{2}` is a negative half-integer,

        .. math::
            \mathcal{I}_{-\bar{m}/2}(s, s_0; a, b)
            \frac{1}{\sqrt{s-s_0}^{\bar{m}}}
            \left(
            \log\left(\frac{\sqrt{s-s_0}-\sqrt{x-s_0}}{\sqrt{s-s_0}+\sqrt{x-s_0}}\right)
            + \sum_{k=1}^{(\bar{m}-1)/2} \frac{2}{2k-1} \frac{\sqrt{s-s_0}^{2k-1}}{\sqrt{x-s_0}^{2k-1}}
            \right)\bigg|_a^b \, .

        """
        return self.__definite_integral__(s, s0, a, b)

    def __indefinite_integral_00__(self, s, s0, x):
        """Special case: n=0."""
        return logC(x - s)

    def __indefinite_integral_ZP__(self, s, s0, x):
        """n is a positive integer."""
        s_shifted = s - s0
        x_shifted = x - s0
        w = s_shifted / x_shifted

        result = np.log(x - s)
        for k in range(1, self.n + 1):
            result += 1 / k * (1 / w) ** k

        return s_shifted**self.n * result

    def __indefinite_integral_ZM__(self, s, s0, x):
        """n is a negative integer."""
        s_shifted = s - s0
        x_shifted = x - s0
        w = s_shifted / x_shifted

        # Special case:
        if self.n == -1:
            return 1 / s_shifted * logC(1 - w)

        # The rest:
        n_abs = -self.n
        result = logC(1 - w)
        for k in range(1, n_abs):
            result += 1 / k * w**k

        return s_shifted**self.n * result

    def __indefinite_integral_ZP_HALF__(self, s, s0, x):
        """n is a positive half-integer."""
        s_shifted = s - s0
        x_shifted = x - s0
        w = s_shifted / x_shifted
        w_sqrt = np.sqrt(w + 0j)

        m = 2 * self.n
        result = logC((w_sqrt - 1) / (w_sqrt + 1))
        for k in range(1, (m + 1) // 2 + 1):
            result += (2 / (2 * k - 1)) * (1 / w_sqrt) ** (2 * k - 1)

        return s_shifted**self.n * result

    def __indefinite_integral_ZM_HALF__(self, s, s0, x):
        """n is a negative half-integer."""
        s_shifted = s - s0
        x_shifted = x - s0
        w = s_shifted / x_shifted
        w_sqrt = np.sqrt(w + 0j)

        m = 2 * self.n
        result = logC((w_sqrt - 1) / (w_sqrt + 1))
        for k in range(1, (m - 1) // 2 + 1):
            result += (2 / (2 * k - 1)) * w_sqrt ** (2 * k - 1)

        return s_shifted**self.n * result


class DispersionIntegralRHC:
    r"""
    Dispersion integral along the right-hand cut.

    Parameters
    ----------
    integrand : callable
        Integrand along the RHC.
    threshold : float
        Lower boundary of the integral.
    infinity : float, optional
        Upper boundary of the integral in the units of :code:`threshold`.
    integration_split_points : array_like, optional
        Integration split points in the units of :code:`threshold`.
    integration_order : int, optional
        Gauss–Legendre quadrature order.
    subtraction_point : float, optional
        Point :math:`s_0` of subtraction. Needs to be below the threshold.
    subtraction_constants : array_like, optional
        Array of :math:`f_i` subtraction constants. Defines subtraction level :math:`n`.

    Notes
    -----
    The integral is defined as

    .. math:: F(s) = \sum_{i=0}^{n-1}f_i (s-s_0)^i + \frac{(s-s_0)^n}{\pi}\int_{s_{thr}}^{\infty}\frac{f(x) \, dx}{(x-s_0)^n(x-s-i\epsilon)},

    where :math:`s_0` is the subtraction point, :math:`f_i` are the subtraction constants,
    :math:`n` is the subtraction level, and
    :math:`f` is the `integrand`.
    """
    def __init__(
        self,
        integrand : callable,
        threshold : float,
        infinity : float = 1e6,
        integration_split_points : list[float] = [2, 5, 10, 50, 100],
        integration_order : int = 100,
        subtraction_point : float = 0,
        subtraction_constants : list[float] = [0],
    ):
        self.integrand = integrand
        self.threshold = threshold
        self.infinity = infinity
        self.interval_split = threshold * np.array(
            [1, *integration_split_points, infinity]
        )

        self.order = integration_order

        self.s0 = subtraction_point
        if self.s0 >= self.threshold:
            warnings.warn(f"Subtraction point s0={self.s0} needs to be below the threshold={self.threshold}.")
        self.f0_arr = subtraction_constants
        self.n_subtr = len(subtraction_constants)

        self.analytic_remainder = SimplePoleMonomial(-self.n_subtr)

        self.__precompute__()
        # self.compute = np.vectorize(self.__compute__)

    def __precompute__(self):
        (self.gl_roots, self.gl_weights) = np.polynomial.legendre.leggauss(self.order)

        intervals = []
        scalings = []
        s_prime = []

        for i in range(len(self.interval_split) - 1):
            a = self.interval_split[i]
            b = self.interval_split[i + 1]
            intervals.append([a, b])
            scalings.append((b - a) / 2)
            s_prime.append(
                np.array([((x + 1) / 2) * (b - a) + a for x in self.gl_roots])
            )

        self.intervals = np.array(intervals)
        self.scalings = np.array(scalings)
        self.scaled_weights = self.gl_weights[:, None] * self.scalings[None, :]
        self.s_prime = np.array(s_prime).T
        self.f_at_s_prime = self.integrand(self.s_prime)

    def __compute_real__(self, s):
        s = np.array(s)
        dim_s = len(s.shape)
        s_axes = tuple(np.arange(dim_s))
        gl_axes = tuple(dim_s + np.arange(2))

        scaled_weights = np.expand_dims(self.scaled_weights, s_axes)
        __s__ = np.expand_dims(s, gl_axes)
        __s_prime__ = np.expand_dims(self.s_prime, s_axes)
        __f_at_s_prime__ = np.expand_dims(self.f_at_s_prime, s_axes)

        f_at_s = self.integrand(s)
        __f_at_s__ = np.expand_dims(f_at_s, gl_axes)

        integrand = (
            (__s__ - self.s0) ** self.n_subtr
            * (__f_at_s_prime__ - __f_at_s__)
            / ((__s_prime__ - self.s0) ** self.n_subtr * (__s_prime__ - __s__))
        )

        numerical_integral = np.zeros(s.shape, dtype=np.complex128)
        analytic_integral = np.zeros(s.shape, dtype=np.complex128)

        numerical_integral += np.sum(integrand * scaled_weights, gl_axes)
        analytic_integral += (
            f_at_s
            * (s - self.s0) ** self.n_subtr
            * np.where(
                np.imag(s) == 0,
                self.analytic_remainder(s, self.s0, self.threshold, np.inf),
                0,
            )
        )

        polynomial_head = 0.0
        for i in range(self.n_subtr):
            polynomial_head += self.f0_arr[i] * (s - self.s0) ** i

        return polynomial_head + (1 / np.pi) * (numerical_integral + analytic_integral)
    
    def __compute_complex__(self, s):
        s = np.array(s)
        dim_s = len(s.shape)
        s_axes = tuple(np.arange(dim_s))
        gl_axes = tuple(dim_s + np.arange(2))

        scaled_weights = np.expand_dims(self.scaled_weights, s_axes)
        __s__ = np.expand_dims(s, gl_axes)
        __s_prime__ = np.expand_dims(self.s_prime, s_axes)
        __f_at_s_prime__ = np.expand_dims(self.f_at_s_prime, s_axes)

        integrand = (
            (__s__ - self.s0) ** self.n_subtr
            * (__f_at_s_prime__)
            / ((__s_prime__ - self.s0) ** self.n_subtr * (__s_prime__ - __s__))
        )

        numerical_integral = np.zeros(s.shape, dtype=np.complex128)
        
        numerical_integral += np.sum(integrand * scaled_weights, gl_axes)

        polynomial_head = 0.0
        for i in range(self.n_subtr):
            polynomial_head += self.f0_arr[i] * (s - self.s0) ** i

        return polynomial_head + (1 / np.pi) * numerical_integral

    def __call__(self, s):
        r"""
        Parameters
        ----------
        s : array_like
            Four-momentum squared of the propagating particle.

        Returns
        -------
        F : array_like
            The same shape as input `s`.
        """
        s = np.array(s)
        is_complex = np.imag(s) != 0
        result = np.zeros(s.shape, dtype=np.complex128)
        if np.sum(is_complex) > 0:
            result[is_complex] = self.__compute_complex__(s[is_complex])
        if np.sum(~is_complex) > 0:
            result[~is_complex] = self.__compute_real__(np.real(s[~is_complex]))
        return result


class OmnesFunction(DispersionIntegralRHC):
    r"""
    Omnès–Muskhelishvili solution :cite:`Omnes:1958hv`.

    Parameters
    ----------
    integrand : callable
        Integrand along the RHC.
    threshold : float
        Lower boundary of the integral.
    infinity : float, optional
        Upper boundary of the integral in the units of :code:`threshold`.
    integration_split_points : array_like, optional
        Integration split points in the units of :code:`threshold`.
    integration_order : int, optional
        Gauss–Legendre quadrature order.
    subtraction_point : float, optional
        Point :math:`s_0` of subtraction.
    subtraction_constants : array_like, optional
        Array of :math:`f_i` subtraction constants. Defines subtraction level :math:`n`.

    Notes
    -----
    The function is defined as

    .. math:: \Omega(s) = \exp\left(\sum_{i=0}^{n-1}f_i (s-s_0)^i + \frac{(s-s_0)^n}{\pi}\int_{s_{thr}}^{\infty}\frac{\delta(x) \, dx}{(x-s_0)^n(x-s-i\epsilon)}\right),

    where :math:`s_0` is the subtraction point, :math:`f_i` are the subtraction constants for :math:`\log\Omega`,
    :math:`n` is the subtraction level, and
    :math:`\delta` is the input phase, passed as `integrand`.

    See Also
    --------
    dispersionrelations.integrals.DispersionIntegralRHC
    """
    def __call__(self, s):
        r"""
        Parameters
        ----------
        s : array_like
            Four-momentum squared of the propagating particle.

        Returns
        -------
        Ω : array_like
            The same shape as input `s`.
        """
        return np.exp(super().__call__(s))



import numpy as np


def Kallen(x, y, z):
    r"""
    The Källén triangle function.

    Parameters
    ----------
    x, y, z : array_like
        A complex number or sequence of complex numbers.

    Returns
    -------
    λ : array_like
        The broadcasted shape from `x`, `y`, `z`.

    Notes
    -----

    The function is defined as :cite:`Kallen:1964lxa`

    .. math:: \lambda(x, y, z) = x^2 + y^2 + z^2 - 2 (xy + yz + zx) .

    An alternate form can be derived for squared inputs

    .. math:: \lambda(q^2, M_1^2, M_2^2) = (q^2 - (M_1 + M_2)^2)(q^2 - (M_1 - M_2)^2) .
    """
    return x**2 + y**2 + z**2 - 2 * (x * y + y * z + z * x)


def momentum_cms(s, M1, M2):
    r"""
    The center-of-mass momentum of a two-body system.

    Parameters
    ----------
    s : array_like
        Four-momentum squared of the two-body system.
    M1 : float, complex
        Mass of the first particle.
    M2 : float, complex
        Mass of the second particle.

    Returns
    -------
    q : array_like
        The same shape as input `s`.

    Notes
    -----
    The CMS momentum is defined as

    .. math:: q(s, M_1, M_2) = \frac{1}{2}\sqrt{\frac{\lambda(s, M_1^2, M_2^2)}{s}},

    where :math:`\lambda(s, M_1^2, M_2^2)` is the Källén triangle function.

    Examples
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from dispersionrelations.kinematics import momentum_cms
    >>> M1 = 0.3
    >>> M2 = 0.5
    >>> E_plot = np.linspace(M1+M2, 2, 500)
    >>> s_plot = E_plot**2
    >>> plt.plot(E_plot, momentum_cms(s_plot, M1, M2))

    .. plot::

        import numpy as np
        import matplotlib.pyplot as plt
        from dispersionrelations.kinematics import momentum_cms
        M1 = 0.3
        M2 = 0.5
        E_plot = np.linspace(M1+M2, 2, 500)
        s_plot = E_plot**2
        plt.plot(E_plot, momentum_cms(s_plot, M1, M2))

    See Also
    --------
    dispersionrelations.kinematics.Kallen
    """
    return np.sqrt(Kallen(s, M1**2, M2**2) / s + 0j) / 2


def phase_space_twobody(s, M1, M2):
    r"""
    Phase space function for a two-particle system.

    Parameters
    ----------
    s : array_like
        Four-momentum squared of the two-body system.
    M1 : float, complex
        Mass of the first particle.
    M2 : float, complex
        Mass of the second particle.

    Returns
    -------
    ρ : array_like
        The same shape as input `s`.

    Notes
    -----
    The phase space function is defined as

    .. math:: \rho(s, M_1, M_2) = \frac{1}{16\pi}\frac{2q}{\sqrt{s}} = \frac{1}{16\pi}\frac{\sqrt{\lambda(s, M_1, M_2)}}{s},

    where :math:`q` is the CMS momentum of the particles.

    Examples
    -------
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> from dispersionrelations.kinematics import phase_space_twobody
    >>> M1 = 0.3
    >>> M2 = 0.5
    >>> E_plot = np.linspace(M1+M2, 2, 500)
    >>> s_plot = E_plot**2
    >>> plt.plot(E_plot, phase_space_twobody(s_plot, M1, M2))

    .. plot::

        import numpy as np
        import matplotlib.pyplot as plt
        from dispersionrelations.kinematics import phase_space_twobody
        M1 = 0.3
        M2 = 0.5
        E_plot = np.linspace(M1+M2, 2, 500)
        s_plot = E_plot**2
        plt.plot(E_plot, phase_space_twobody(s_plot, M1, M2))

    See Also
    --------
    dispersionrelations.kinematics.momentum_cms, dispersionrelations.kinematics.Kallen
    """
    return 1 / (16 * np.pi) * np.sqrt(Kallen(s, M1**2, M2**2) / s**2 + 0j)



r"""
Physical and mathematical constants and related functions used throughout the package.
Most masses and widths are taken by the Review of Particle Physics by the Particle Data Group :cite:`Zyla:2020zbs`.
The base unit is taken to be `GeV`.
"""

import numpy as np
from uncertainties import ufloat


MeV = 1e-3
r"""Megaelectronvolt, numerically defined as :math:`10^{-3}` as a default (so that :code:`GeV` :math:`=1`)."""

eV = 1e-6 * MeV
r"""Electronvolt, numerically defined as :math:`10^{-6}\times` :code:`MeV`."""

keV = 1e-3 * MeV
r"""Kiloelectronvolt, numerically defined as :math:`10^{-3}\times` :code:`MeV`."""

GeV = 1e3 * MeV
r"""Gigaelectronvolt, numerically defined as :math:`10^{3}\times` :code:`MeV`."""

TeV = 1e6 * GeV
r"""Teraelectronvolt, numerically defined as :math:`10^{6}\times` :code:`MeV`."""

E_almost_zero = 1e-5 * GeV
r"""A small energy, used in order to avoid division by zero."""

s_almost_zero = E_almost_zero**2

alpha_fs = 1 / 137.036
r"""Electromagnetic fine structure constant :math:`\alpha_{\text{fs}}\approx1/137`."""

e2 = 4 * np.pi * alpha_fs
r"""Electric charge squared, :math:`e^2 = 4\pi\alpha_{\text{fs}}`."""

to_nb = 1 / (2.56819 * 1e-6) * GeV**2
r"""Conversion factor from natural units to nanobarns.

**Note**: a "natural unit" is defined by whatever is equal to `1` by default.
If :math:`\text{GeV}=1`, then the constant :code:`to_nb` will convert
:math:`{\text{GeV}}^{-2}` to nanobarns."""

degrees = np.pi / 180
r"""Conversion factor from radians to degrees :math:`=\pi/180`.

----"""


M_PI = 139.57 * MeV
r"""Mass of the charged pion :cite:`Zyla:2020zbs`."""

M_PI0 = 134.98 * MeV
r"""Mass of the neutral pion :cite:`Zyla:2020zbs`."""

M_K = 493.677 * MeV
r"""Mass of the charged kaon :cite:`Zyla:2020zbs`."""

M_K0 = 497.611 * MeV
r"""Mass of the neutral kaon :cite:`Zyla:2020zbs`."""

M_RHO = 762.5 * MeV
r"""Mass of the :math:`\rho` meson, obtained from the pole position :cite:`Garcia-Martin:2011nna`."""

M_RHO_STD = 1.7 * MeV  # largest error
r"""Standard deviation of the :math:`\rho` meson mass, obtained from the pole position :cite:`Garcia-Martin:2011nna`."""

G_RHO = 2 * 73.2 * MeV
r"""Width of the :math:`\rho` meson, obtained from the pole position :cite:`Garcia-Martin:2011nna`."""

G_RHO_STD = 2 * 1.1 * MeV  # largest error
r"""Standard deviation of the :math:`\rho` meson width, obtained from the pole position :cite:`Garcia-Martin:2011nna`."""

M_RHO_BW = 775.26 * MeV
r"""Mass of the :math:`\rho` meson, obtained as a Breit–Wigner parameter :cite:`Zyla:2020zbs`."""

M_RHO_BW_STD = 0.23 * MeV
r"""Standard deviation of the :math:`\rho` meson mass, obtained as a Breit–Wigner parameter :cite:`Zyla:2020zbs`."""

G_RHO_BW = 147.4 * MeV
r"""Width of the :math:`\rho` meson, obtained as a Breit–Wigner parameter :cite:`Zyla:2020zbs`."""

G_RHO_BW_STD = 0.8 * MeV
r"""Standard deviation of the :math:`\rho` meson width, obtained as a Breit–Wigner parameter :cite:`Zyla:2020zbs`."""

M_OMEGA = 782.66 * MeV
r"""Mass of the :math:`\omega` meson :cite:`Zyla:2020zbs`."""

M_OMEGA_STD = 0.13 * MeV  # largest error
r"""Standard deviation of the :math:`\omega` meson mass :cite:`Zyla:2020zbs`."""

G_OMEGA = 8.68 * MeV
r"""Width of the :math:`\omega` meson :cite:`Zyla:2020zbs`."""

G_OMEGA_STD = 0.13 * MeV  # largest error
r"""Standard deviation of the :math:`\omega` meson width :cite:`Zyla:2020zbs`."""

M_PHI = 1019.461 * MeV
r"""Mass of the :math:`\phi` meson :cite:`Zyla:2020zbs`."""

M_PHI_STD = 0.016 * MeV
r"""Standard deviation of the :math:`\phi` meson mass :cite:`Zyla:2020zbs`."""

G_PHI = 4.25 * MeV
r"""Width of the :math:`\phi` meson :cite:`Zyla:2020zbs`."""

G_PHI_STD = 0.013 * MeV
r"""Standard deviation of the :math:`\phi` meson width :cite:`Zyla:2020zbs`."""

M_ETA = 547.862 * MeV
r"""Mass of the :math:`\eta` meson :cite:`Zyla:2020zbs`."""

M_ETA_STD = 0.017 * MeV
r"""Standard deviation of the :math:`\eta` meson mass :cite:`Zyla:2020zbs`."""

G_ETA = 1.31 * keV
r"""Width of the :math:`\eta` meson :cite:`Zyla:2020zbs`."""

G_ETA_STD = 0.05 * keV
r"""Standard deviation of the :math:`\eta` meson width :cite:`Zyla:2020zbs`."""

M_KSTAR = 890 * MeV
r"""Mass of the :math:`K^*` :cite:`Zyla:2020zbs`."""

M_KSTAR_STD = 14 * MeV
r"""Standard deviation of the :math:`K^*` mass :cite:`Zyla:2020zbs`."""

G_KSTAR = 2 * 26 * MeV
r"""Width of the :math:`K^*` :cite:`Zyla:2020zbs`."""

G_KSTAR_STD = 2 * 6 * MeV
r"""Standard deviation of the :math:`K^*` width :cite:`Zyla:2020zbs`."""

M_A1_POLE = 1209 * MeV
r"""Mass of the :math:`a_1`, obtained from the pole position :cite:`JPAC:2018zwp`."""

M_A1_POLE_STD = 16 * MeV  # largest error
r"""Standard deviation of the :math:`a_1` mass, obtained from the pole position :cite:`JPAC:2018zwp`."""

G_A1_POLE = 576 * MeV
r"""Width of the :math:`a_1`, obtained from the pole position :cite:`JPAC:2018zwp`."""

G_A1_POLE_STD = 100 * MeV  # largest error
r"""Standard deviation of the :math:`a_1` width, obtained from the pole position :cite:`JPAC:2018zwp`."""

BR_OMEGA_TO_3PI = 89.2 / 100
r"""Branching ratio of :math:`\omega \to 3\pi` decay :cite:`Zyla:2020zbs`."""

BR_OMEGA_TO_3PI_STD = 0.7 / 100
r"""Standard deviation of the branching ratio of :math:`\omega \to 3\pi` decay :cite:`Zyla:2020zbs`."""

__G_OMEGA = ufloat(G_OMEGA, G_OMEGA_STD)
__BR_OMEGA_TO_3PI = ufloat(BR_OMEGA_TO_3PI, BR_OMEGA_TO_3PI_STD)
__G_OMEGA_TO_3PI = __G_OMEGA * __BR_OMEGA_TO_3PI

G_OMEGA_TO_3PI = float(__G_OMEGA_TO_3PI.nominal_value)
r"""Partial width of :math:`\omega \to 3\pi` decay."""

G_OMEGA_TO_3PI_STD = float(__G_OMEGA_TO_3PI.std_dev)
r"""Standard deviation of the partial width of :math:`\omega \to 3\pi` decay."""

BR_OMEGA_TO_PI0GAMMA = 8.33 / 100
r"""Branching ratio of :math:`\omega \to \pi^0 \gamma` decay :cite:`Zyla:2020zbs`."""

BR_OMEGA_TO_PI0GAMMA_STD = 0.25 / 100
r"""Standard deviation of the branching ratio of :math:`\omega \to \pi^0 \gamma` decay :cite:`Zyla:2020zbs`.""" 

__BR_OMEGA_TO_PI0GAMMA = ufloat(BR_OMEGA_TO_PI0GAMMA, BR_OMEGA_TO_PI0GAMMA_STD)
__G_OMEGA_TO_PI0GAMMA = __G_OMEGA * __BR_OMEGA_TO_PI0GAMMA

G_OMEGA_TO_PI0GAMMA = float(__G_OMEGA_TO_PI0GAMMA.nominal_value)
r"""Partial width of :math:`\omega \to \pi^0 \gamma` decay."""

G_OMEGA_TO_PI0GAMMA_STD = float(__G_OMEGA_TO_PI0GAMMA.std_dev)
r"""Standard deviation of the partial width of :math:`\omega \to \pi^0 \gamma` decay."""

G_RHO_TO_ETAGAMMA = 44.2 * keV
r"""Partial width of :math:`\rho \to \eta \gamma` decay :cite:`Schafer:2023qtl`."""

G_RHO_TO_ETAGAMMA_STD = 3.1 * keV
r"""Standard deviation of the partial width of :math:`\rho \to \eta \gamma` decay :cite:`Schafer:2023qtl`."""

BR_KSTAR_TO_KGAMMA_CHARGED = 0.098 / 100
r"""Branching ratio of the charged :math:`K^* \to K \gamma` decay :cite:`Dax:2020dzg`."""

BR_KSTAR_TO_KGAMMA_CHARGED_STD = 0.009 / 100
r"""Standard deviation of the branching ratio of the charged :math:`K^* \to K \gamma` decay :cite:`Dax:2020dzg`."""

BR_KSTAR_TO_KGAMMA_NEUTRAL = 0.246 / 100
r"""Branching ratio of the neutral :math:`K^* \to K \gamma` decay :cite:`Dax:2020dzg`."""

BR_KSTAR_TO_KGAMMA_NEUTRAL_STD = 0.021 / 100
r"""Standard deviation of the branching ratio of the neutral :math:`K^* \to K \gamma` decay :cite:`Dax:2020dzg`."""

__G_KSTAR = ufloat(G_KSTAR, G_KSTAR_STD)
__BR_KSTAR_TO_KGAMMA_CHARGED = ufloat(BR_KSTAR_TO_KGAMMA_CHARGED, BR_KSTAR_TO_KGAMMA_CHARGED_STD)
__BR_KSTAR_TO_KGAMMA_NEUTRAL = ufloat(BR_KSTAR_TO_KGAMMA_NEUTRAL, BR_KSTAR_TO_KGAMMA_NEUTRAL_STD)
__G_KSTAR_TO_KGAMMA_CHARGED = __G_KSTAR * __BR_KSTAR_TO_KGAMMA_CHARGED
__G_KSTAR_TO_KGAMMA_NEUTRAL = __G_KSTAR * __BR_KSTAR_TO_KGAMMA_NEUTRAL

G_KSTAR_TO_KGAMMA_CHARGED = float(__G_KSTAR_TO_KGAMMA_CHARGED.nominal_value)
r"""Partial width of the charged :math:`K^* \to K \gamma` decay."""

G_KSTAR_TO_KGAMMA_CHARGED_STD = float(__G_KSTAR_TO_KGAMMA_CHARGED.std_dev)
r"""Standard deviation of the partial width of the charged :math:`K^* \to K \gamma` decay."""

G_KSTAR_TO_KGAMMA_NEUTRAL = float(__G_KSTAR_TO_KGAMMA_NEUTRAL.nominal_value)
r"""Partial width of the neutral :math:`K^* \to K \gamma` decay."""

G_KSTAR_TO_KGAMMA_NEUTRAL_STD = float(__G_KSTAR_TO_KGAMMA_NEUTRAL.std_dev)
r"""Standard deviation of the partial width of the neutral :math:`K^* \to K \gamma` decay."""


import numpy as np
import scipy as sp
import warnings
import os

from dispersionrelations.utils import cite, uncite


def check_pd(mat):
    r"""
    Check if a matrix is positive definite.

    Parameters
    ----------
    mat : array-like, shape (N, N)
        The matrix to be checked.

    Returns
    -------
    bool
        True if the matrix is positive definite, False otherwise.
    """
    try:
        np.linalg.cholesky(mat)
        return True
    except np.linalg.LinAlgError:
        return False


def corr_to_cov(corr, std):
    r"""
    Convert a correlation matrix to a covariance matrix.

    Parameters
    ----------
    corr : array-like, shape (N, N)
        The correlation matrix to be converted.
    std : array-like, shape (N,)
        The standard deviations corresponding to the variables.

    Returns
    -------
    cov : array-like, shape (N, N)
        The resulting covariance matrix.

    Raises
    ------
    ValueError
        If corr is not a square matrix or if corr and std have incompatible dimensions.
    """
    corr = np.asarray(corr)
    std = np.asarray(std)
    if corr.shape[0] != corr.shape[1]:
        raise ValueError(f"corr must be a square matrix, but got shape {corr.shape}")
    if corr.shape[0] != std.shape[0]:
        raise ValueError(
            f"corr and std must have compatible dimensions, but got {corr.shape} and {std.shape}"
        )
    if std.ndim != 1:
        raise ValueError(
            f"std must be a one-dimensional array, but got shape {std.shape}"
        )
    outer = np.outer(std, std)
    cov = corr * outer
    return cov


def cov_to_corr(cov):
    r"""
    Convert a covariance matrix to a correlation matrix.

    Parameters
    ----------
    cov : array-like, shape (N, N)
        The covariance matrix to be converted.

    Returns
    -------
    corr : array-like, shape (N, N)
        The resulting correlation matrix.

    Raises
    ------
    ValueError
        If cov is not a square matrix.
    """
    cov = np.asarray(cov)
    if cov.shape[0] != cov.shape[1]:
        raise ValueError(f"cov must be a square matrix, but got shape {cov.shape}")
    corr = np.zeros(cov.shape)
    std = np.sqrt(np.diag(cov))
    outer = np.outer(std, std)
    nonzero = np.where(outer != 0)
    corr[nonzero] = cov[nonzero] / outer[nonzero]
    corr[np.isnan(corr)] = 0
    return corr


def cov_to_std(cov):
    r"""
    Convert a covariance matrix to standard deviations (square root of the diagonal elements).

    Parameters
    ----------
    cov : array-like, shape (N, N)
        The covariance matrix to be converted.

    Returns
    -------
    std : array-like, shape (N,)
        The resulting standard deviations.

    Raises
    ------
    ValueError
        If cov is not a square matrix.
    """
    cov = np.asarray(cov)
    if cov.shape[0] != cov.shape[1]:
        raise ValueError(f"cov must be a square matrix, but got shape {cov.shape}")
    return np.sqrt(np.diag(cov))


def std_to_cov(std, correlation=0.0):
    r"""
    Convert standard deviations to a covariance matrix, assuming a given correlation structure.

    Parameters
    ----------
    std : array-like, shape (N,)
        The standard deviations to be converted.
    correlation : float, optional
        The correlation coefficient to be used in the conversion. Default is 0.0 (no correlation).

    Returns
    -------
    cov : array-like, shape (N, N)
        The resulting covariance matrix.

    Raises
    ------
    ValueError
        If std is not a one-dimensional array.
    """
    std = np.asarray(std)
    if std.ndim != 1:
        raise ValueError(
            f"std must be a one-dimensional array, but got shape {std.shape}"
        )
    corr = (1.0 - correlation) * np.eye(len(std)) + correlation * np.ones(
        (len(std), len(std))
    )
    cov = np.outer(std, std) * corr
    return cov


def chi2(fit_data, observed_data, covariance_matrix, weights=None):
    r"""
    :math:`\chi^2`, computed using modelled and observed data.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    covariance_matrix : array_like
        Covariance matrix of the experimental data, :math:`C \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.

    Returns
    -------
    s : float
        The total :math:`\chi^2`.

    Notes
    -----
    The :math:`\chi^2` is calculated using

    .. math:: \chi^2 = \sum_i \sum_j (f_i-o_i)(C^{-1}_{ij}W_{ij})(f_j-o_j).

    See also
    --------
    dispersionrelations.data.chi2_vector : for individual contributions.
    """
    if weights is not None:
        covariance_matrix = covariance_matrix / weights
    difference = fit_data - observed_data
    return difference @ np.linalg.inv(covariance_matrix) @ difference


def chi2_with_inverse(fit_data, observed_data, inverse_covariance_matrix, weights=None):
    r"""
    :math:`\chi^2`, computed using modelled and observed data.
    This version of the function accepts the inverse of the covariance matrix
    as a parameter and therefore gains speed by avoiding matrix inversion.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    inverse_covariance_matrix : array_like
        Inverted covariance matrix of the experimental data, :math:`I \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.

    Returns
    -------
    s : float
        The total :math:`\chi^2`.

    Notes
    -----
    The :math:`\chi^2` is calculated using

    .. math:: \chi^2 = \sum_i \sum_j (f_i-o_i)(I_{ij}W_{ij})(f_j-o_j).

    See also
    --------
    dispersionrelations.data.chi2_vector_with_inverse : for individual contributions.
    """
    if weights is not None:
        inverse_covariance_matrix = inverse_covariance_matrix * weights
    difference = fit_data - observed_data
    return difference @ inverse_covariance_matrix @ difference


def chi2_vector(fit_data, observed_data, covariance_matrix, weights=None):
    r"""
    :math:`\chi^2` vector.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    covariance_matrix : array_like
        Covariance matrix of the experimental data, :math:`C \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.

    Returns
    -------
    v : array_like
        A vector with shape (n,) of individual :math:`\chi^2` contributions.

    Notes
    -----
    The :math:`\chi^2` vector is built using

    .. math:: v_i = \sum_j (f_i-o_i)(C^{-1}_{ij}W_{ij})(f_j-o_j).

    Warning
    -------
    When the covariance matrix is non-diagonal, the individual components of this
    vector do not have a statistical meaning, since the observations are correlated.
    Use with caution!

    Examples
    --------
    >>> import numpy as np
    >>> from dispersionrelations.data import chi2_vector
    >>> np.random.seed(137)
    >>> n = 10
    >>> x_f = np.linspace(1, 2, n)
    >>> x_o = x_f + 0.1 * np.random.rand(n)
    >>> x_e = 0.1 * (np.random.rand(n) + 1) / 2
    >>> x_C = np.diag(x_e)**2
    >>> print(chi2_vector(x_f, x_o, x_C))
    [0.89280398 0.01007333 0.62584069 0.67718574 0.32026806 0.99981995 0.91740446 0.01387227 2.33635045 0.18841755]

    .. plot::

        import numpy as np
        from dispersionrelations.data import chi2_vector
        import matplotlib.pyplot as plt

        np.random.seed(137)
        n = 10
        x_f = np.linspace(1, 2, n)
        x_o = x_f + 0.1 * np.random.rand(n)
        x_e = 0.1 * (np.random.rand(n) + 1) / 2
        x_C = np.diag(x_e)**2
        x_chi2_vector = chi2_vector(x_f, x_o, x_C)
        plt.errorbar(np.arange(n), x_o, yerr=x_e, fmt="o", markersize=0, capsize=2, label="observed_data", zorder=2)
        plt.plot(np.arange(n), x_f, label="fit_data", zorder=2)
        plt.fill_between(np.arange(n), 0.1 * x_chi2_vector, 0, color='tab:red', label=r"$\chi^2/10$", zorder=3)
        plt.legend()

    See also
    --------
    dispersionrelations.data.chi2
    """
    if weights is not None:
        covariance_matrix = covariance_matrix / weights
    difference = fit_data - observed_data
    return difference * (np.linalg.inv(covariance_matrix) @ difference)


def chi2_vector_with_inverse(
    fit_data, observed_data, inverse_covariance_matrix, weights=None
):
    r"""
    :math:`\chi^2`, computed using modelled and observed data.
    This version of the function accepts the inverse of the covariance matrix
    as a parameter and therefore gains speed by avoiding matrix inversion.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    inverse_covariance_matrix : array_like
        Inverted covariance matrix of the experimental data, :math:`I \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.

    Returns
    -------
    v : array_like
        A vector with shape (n,) of individual :math:`\chi^2` contributions.

    Notes
    -----
    The :math:`\chi^2` vector is built using

    .. math:: v_i = \sum_j (f_i-o_i)(I_{ij}W_{ij})(f_j-o_j).

    See also
    --------
    dispersionrelations.data.chi2_with_inverse
    """
    if weights is not None:
        inverse_covariance_matrix = inverse_covariance_matrix * weights
    difference = fit_data - observed_data
    return difference * (inverse_covariance_matrix @ difference)


def chi2_reduced(
    fit_data, observed_data, covariance_matrix, weights=None, number_of_parameters=0
):
    r"""
    A reduced :math:`\chi^2`, computed using modelled and observed data.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    covariance_matrix : array_like
        Covariance matrix of the experimental data, :math:`C \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.
    number_of_parameters : int
        Number of parameters in the model, used for calculating the degrees of freedom.

    Returns
    -------
    s : float
        The reduced :math:`\chi^2`.

    Notes
    -----
    The reduced :math:`\chi^2` is calculated using

    .. math:: \bar{\chi}^2 = \frac{\chi^2}{\text{d.o.f.}} = \frac{1}{n - n_\text{par}} \sum_i \sum_j (f_i-o_i)(C^{-1}_{ij}W_{ij})(f_j-o_j).

    See also
    --------
    dispersionrelations.data.chi2
    """
    degrees_of_freedom = len(observed_data) - number_of_parameters
    return (
        chi2(fit_data, observed_data, covariance_matrix, weights) / degrees_of_freedom
    )


def chi2_reduced_with_inverse(
    fit_data,
    observed_data,
    inverse_covariance_matrix,
    weights=None,
    number_of_parameters=0,
):
    r"""
    A reduced :math:`\chi^2`, computed using modelled and observed data.
    This version of the function accepts the inverse of the covariance matrix
    as a parameter and therefore gains speed by avoiding matrix inversion.

    Parameters
    ----------
    fit_data : array_like
        Data obtained from a model, :math:`f_i \in \{f_1, f_2, \dots, f_n\}`.
    observed_data : array_like
        Data obtained from an experiment, :math:`o_i \in \{o_1, o_2, \dots, o_n\}`.
    inverse_covariance_matrix : array_like
        Inverted covariance matrix of the experimental data, :math:`I \in \mathbb{R}^{n\times n}`.
    weights : array_like
        Weight matrix, used to manually attenuate parts of the data, :math:`W \in \mathbb{R}^{n\times n}`.
    number_of_parameters : int
        Number of parameters in the model, used for calculating the degrees of freedom.

    Returns
    -------
    s : float
        The reduced :math:`\chi^2`.

    Notes
    -----
    The reduced :math:`\chi^2` is calculated using

    .. math:: \bar{\chi}^2 = \frac{\chi^2}{\text{d.o.f.}} = \frac{1}{n - n_\text{par}} \sum_i \sum_j (f_i-o_i)(I_{ij}W_{ij})(f_j-o_j).

    See also
    --------
    dispersionrelations.data.chi2
    """
    degrees_of_freedom = len(observed_data) - number_of_parameters
    return (
        chi2_with_inverse(fit_data, observed_data, inverse_covariance_matrix, weights)
        / degrees_of_freedom
    )


class Datablock:
    r"""
    A class to represent a dataset with x and y values, along with statistical and systematic covariance matrices.

    Parameters
    ----------
    x : array-like, shape (N,)
        The x values of the data points.
    y : array-like, shape (N,)
        The y values of the data points.
    cov_stat : array-like, shape (N, N), optional
        The initial statistical covariance matrix. Default is None, which initializes it to a zero matrix.
    cov_syst : array-like, shape (N, N), optional
        The initial systematic covariance matrix. Default is None, which initializes it to a zero matrix.
    metadata : dict, optional
        A dictionary to store additional metadata about the dataset. Default is an empty dictionary.
        Common keys include 'ref', 'header', 'footer', utilised by the :class:`Datawriter` class.

    Attributes
    ----------
    x : array-like, shape (N,)
        The x values of the data points.
    y : array-like, shape (N,)
        The y values of the data points.
    N : int
        The number of data points.
    cov_stat : array-like, shape (N, N)
        The statistical covariance matrix.
    cov_syst : array-like, shape (N, N)
        The systematic covariance matrix.
    metadata : dict
        A dictionary to store additional metadata about the dataset.
    """

    def __init__(self, x, y, xlabel, ylabel, cov_stat=None, cov_syst=None, metadata={}):
        self.x = x
        self.y = y
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.N = len(x)
        if len(y) != self.N:
            raise ValueError(
                f"x and y must have the same length, but got {len(x)} and {len(y)}"
            )
        self.cov_stat = np.zeros((self.N, self.N))
        self.cov_syst = np.zeros((self.N, self.N))

        if cov_stat is not None:
            self.add_cov_stat(cov_stat)
            self.check_cov_pd()
        if cov_syst is not None:
            self.add_cov_syst(cov_syst)
            self.check_cov_pd()

        self.metadata = metadata

    def check_cov_pd(self):
        r"""
        Check if the total covariance matrix (statistical + systematic) is positive definite.
        Display a warning if it is not.

        Returns
        -------
        None
        """
        if check_pd(self.cov_stat + self.cov_syst) == False:
            warnings.warn("The total covariance matrix is not positive definite")

    def get_std_stat(self):
        r"""
        Get the statistical standard deviations (square root of the diagonal elements of the statistical covariance matrix).

        Returns
        -------
        std_stat : array-like, shape (N,)
            The statistical standard deviations.
        """
        return cov_to_std(self.cov_stat)

    def get_std_syst(self):
        r"""
        Get the systematic standard deviations (square root of the diagonal elements of the systematic covariance matrix).

        Returns
        -------
        std_syst : array-like, shape (N,)
            The systematic standard deviations.
        """
        return cov_to_std(self.cov_syst)

    def get_std_total(self, add_in_quadrature=True):
        r"""
        Get the total standard deviations (square root of the diagonal elements of the total covariance matrix).

        Parameters
        ----------
        add_in_quadrature : bool, optional
            If True, compute the total standard deviations by adding the statistical and systematic standard deviations in quadrature.
            If False, compute the total standard deviations from the total covariance matrix. Default is True.

        Returns
        -------
        std_total : array-like, shape (N,)
            The total standard deviations.
        """
        std_stat = self.get_std_stat()
        std_syst = self.get_std_syst()
        if add_in_quadrature:
            return np.sqrt(std_stat**2 + std_syst**2)
        else:
            return std_stat + std_syst

    def add_cov_stat(self, cov_stat):
        r"""
        Add a contribution to the statistical covariance matrix.

        Parameters
        ----------
        cov_stat : array-like, shape (N, N)
            The statistical covariance matrix to be added.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If cov_stat is not of shape (N, N).
        """
        if cov_stat.shape != (self.N, self.N):
            raise ValueError(
                f"cov_stat must be of shape (N, N), but got {cov_stat.shape}"
            )
        self.cov_stat += cov_stat

    def add_cov_syst(self, cov_syst):
        r"""
        Add a contribution to the systematic covariance matrix.

        Parameters
        ----------
        cov_syst : array-like, shape (N, N)
            The systematic covariance matrix to be added.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If cov_syst is not of shape (N, N).
        """
        if cov_syst.shape != (self.N, self.N):
            raise ValueError(
                f"cov_syst must be of shape (N, N), but got {cov_syst.shape}"
            )
        self.cov_syst += cov_syst

    def add_std_stat(self, std_stat, correlation=0.0):
        r"""
        Add a contribution to the statistical covariance matrix from standard deviations.

        Parameters
        ----------
        std_stat : array-like, shape (N,) or scalar
            The standard deviations to be converted to covariance matrix.
        correlation : float, optional
            The correlation coefficient to be used in the conversion. Default is 0.0 (no correlation).

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If std_stat is not a scalar or of shape (N,).
        """
        std_stat = np.asarray(std_stat)
        if std_stat.shape != (self.N,) and std_stat.shape != (1,):
            raise ValueError(
                f"std_stat must be a scalar or of shape (N,), but got {std_stat.shape}"
            )
        cov_stat = std_to_cov(std_stat, correlation=correlation)
        self.add_cov_stat(cov_stat)

    def add_rel_stat(self, rel_stat, correlation=0.0):
        r"""
        Add a contribution to the statistical covariance matrix from relative uncertainties.

        Parameters
        ----------
        rel_stat : array-like, shape (N,) or scalar
            The relative uncertainties to be converted to standard deviations.
        correlation : float, optional
            The correlation coefficient to be used in the conversion. Default is 0.0 (no correlation).

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If rel_stat is not a scalar or of shape (N,).
        """
        std_stat = np.abs(rel_stat * self.y)
        self.add_std_stat(std_stat, correlation=correlation)

    def add_std_syst(self, std_syst, correlation=1.0):
        r"""
        Add a contribution to the systematic covariance matrix from standard deviations.

        Parameters
        ----------
        std_syst : array-like, shape (N,) or scalar
            The standard deviations to be converted to covariance matrix.
        correlation : float, optional
            The correlation coefficient to be used in the conversion. Default is 1.0 (full correlation).

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If std_syst is not a scalar or of shape (N,).
        """
        std_syst = np.asarray(std_syst)
        if std_syst.shape != (self.N,) and std_syst.shape != (1,):
            raise ValueError(
                f"std_syst must be a scalar or of shape (N,), but got {std_syst.shape}"
            )
        cov_syst = std_to_cov(std_syst, correlation=correlation)
        self.add_cov_syst(cov_syst)

    def add_rel_syst(self, rel_syst, correlation=1.0):
        r"""
        Add a contribution to the systematic covariance matrix from relative uncertainties.

        Parameters
        ----------
        rel_syst : array-like, shape (N,) or scalar
            The relative uncertainties to be converted to standard deviations.
        correlation : float, optional
            The correlation coefficient to be used in the conversion. Default is 1.0 (full correlation).

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If rel_syst is not a scalar or of shape (N,).
        """
        std_syst = np.abs(rel_syst * self.y)
        self.add_std_syst(std_syst, correlation=correlation)

    def add_external_calibration(self, c, c_std, correlation=1.0, new_ylabel=None):
        r"""
        Apply an external calibration factor to the data and propagate its uncertainty.

        Parameters
        ----------
        c : float
            The calibration factor to be applied to the data.
        c_std : float
            The uncertainty of the calibration factor.
        correlation : float, optional
            The correlation coefficient to be used in the propagation of the calibration uncertainty. Default is 1.0 (full correlation).
        new_ylabel : str, optional
            The new label for the y-axis after calibration. If None, the label remains unchanged.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If c or c_std is not a scalar.

        Notes
        -----
        For x-dependent calibration factors, use `rescale_y` instead.
        """
        if np.asarray(c).shape != () or np.asarray(c_std).shape != ():
            raise ValueError("c and c_std must be scalars")
        self.y = c * self.y
        self.cov_stat *= c**2
        self.cov_syst *= c**2
        if new_ylabel is not None:
            self.ylabel = new_ylabel
        if c_std != 0:
            self.add_rel_syst(c_std / c, correlation=correlation)
            self.check_cov_pd()

    def rescale_y(
        self,
        factor,
        factor_std=None,
        stat_or_syst="syst",
        correlation=0.0,
        new_ylabel=None,
    ):
        r"""
        Rescale the y values by a given factor without propagating the uncertainty.

        Parameters
        ----------
        factor : array-like, shape (N,)
            The factor by which to rescale the y values. If an array is provided, it must be of shape (N,).
        factor_std : array-like, shape (N,)
            The uncertainty of the rescaling factor (not propagated).
        stat_or_syst : str, optional
            Specify whether the uncertainty is 'stat' (statistical) or 'syst' (systematic). Default is 'syst'.
        correlation : float, optional
            The correlation coefficient to be used for the uncertainty propagation. Default is 0.0 (no correlation).
        new_ylabel : str, optional
            The new label for the y-axis after rescaling. If None, the label remains unchanged

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If factor is not a scalar or of shape (N,).

        Notes
        -----
        For a global rescaling, use `add_external_calibration` instead.
        """
        factor = np.asarray(factor)
        factor_std = (
            np.asarray(factor_std) if factor_std is not None else np.zeros(factor.shape)
        )
        if factor.shape != (self.N,):
            raise ValueError(
                f"factor must be a scalar or of shape (N,), but got {factor.shape}"
            )
        if factor_std.shape != (self.N,):
            raise ValueError(
                f"factor_std must be a scalar or of shape (N,), but got {factor_std.shape}"
            )
        if stat_or_syst not in ["stat", "syst"]:
            raise ValueError("stat_or_syst must be 'stat' or 'syst'")
        self.y = factor * self.y
        factor_mat = np.outer(factor, factor)
        self.cov_stat = factor_mat * self.cov_stat
        self.cov_syst = factor_mat * self.cov_syst
        if new_ylabel is not None:
            self.ylabel = new_ylabel
        if np.any(factor_std != 0):
            rel_syst = np.abs(factor_std / factor)
            if stat_or_syst == "stat":
                self.add_rel_stat(rel_syst, correlation=correlation)
            elif stat_or_syst == "syst":
                self.add_rel_syst(rel_syst, correlation=correlation)
            self.check_cov_pd()

    def filter_data(self, mask):
        r"""
        Filter the data based on a boolean mask.

        Parameters
        ----------
        mask : array-like, shape (N,)
            A boolean array indicating which data points to keep.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If mask is not of length N.
        """
        mask = np.asarray(mask)
        if mask.shape != (self.N,):
            raise ValueError(
                f"The mask must be of length N={self.N}, but got {mask.shape}"
            )
        self.x = self.x[mask]
        self.y = self.y[mask]
        self.cov_stat = self.cov_stat[np.ix_(mask, mask)]
        self.cov_syst = self.cov_syst[np.ix_(mask, mask)]
        self.N = len(self.x)
        self.check_cov_pd()

    def filter_xrange(self, x_min=None, x_max=None):
        r"""
        Filter the data based on x range.

        Parameters
        ----------
        x_min : float, optional
            The minimum x value to keep. If None, no lower bound is applied.
        x_max : float, optional
            The maximum x value to keep. If None, no upper bound is applied.

        Returns
        -------
        None
        """
        mask_lower = (
            self.x >= x_min if x_min is not None else np.ones(self.N, dtype=bool)
        )
        mask_upper = (
            self.x <= x_max if x_max is not None else np.ones(self.N, dtype=bool)
        )
        self.filter_data(mask_lower * mask_upper)

    def filter_yrange(self, y_min=None, y_max=None):
        r"""
        Filter the data based on y range.

        Parameters
        ----------
        y_min : float, optional
            The minimum y value to keep. If None, no lower bound is applied.
        y_max : float, optional
            The maximum y value to keep. If None, no upper bound is applied.

        Returns
        -------
        None
        """
        mask_lower = (
            self.y >= y_min if y_min is not None else np.ones(self.N, dtype=bool)
        )
        mask_upper = (
            self.y <= y_max if y_max is not None else np.ones(self.N, dtype=bool)
        )
        self.filter_data(mask_lower * mask_upper)

    def remove_datapoints(self, indices):
        r"""
        Remove data points based on their indices.

        Parameters
        ----------
        indices : array-like, shape (M,)
            An array of indices of the data points to be removed.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If any index is out of bounds.
        """
        indices = np.asarray(indices)
        if np.any(indices < 0) or np.any(indices >= self.N):
            raise ValueError(
                f"All indices must be in the range [0, {self.N}), but got {indices}"
            )
        mask = np.ones(self.N, dtype=bool)
        mask[indices] = False
        self.filter_data(mask)

    def downsample_by(self, factor):
        r"""
        Downsample the data by a given factor.

        Parameters
        ----------
        factor : int
            The downsampling factor. Must be a positive integer.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If factor is not a positive integer.
        """
        if not isinstance(factor, int) or factor <= 0:
            raise ValueError("factor must be a positive integer")
        mask = np.zeros(self.N, dtype=bool)
        mask[::factor] = True
        self.filter_data(mask)

    def downsample_to(self, N_points):
        r"""
        Downsample the data to a given number of points.

        Parameters
        ----------
        N_points : int
            The desired number of data points after downsampling. Must be a positive integer less than or equal to N.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If N_points is not a positive integer or greater than N.
        """
        raise NotImplementedError("downsample_to is not implemented yet")

    def interpolate(self, new_x):
        r"""
        Interpolate the data to new x values using linear interpolation.

        Parameters
        ----------
        new_x : array-like, shape (M,)
            The new x values to interpolate to.

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If new_x is not a one-dimensional array.
        """
        raise NotImplementedError("interpolate is not implemented yet")

    def __str__(self):
        result = f"Datablock with {self.N} points(s): {self.xlabel} vs {self.ylabel}."
        if "ref" in self.metadata:
            result += f" Reference: {self.metadata['ref']}."
        return result

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        r"""
        Add two Datablock objects.

        Parameters
        ----------
        other : Datablock
            The other Datablock to be added.

        Returns
        -------
        Datablock
            A new Datablock containing the combined data.

        Raises
        ------
        ValueError
            If other is not a Datablock.

        Notes
        -----
        Metadata and labels are not fully preserved in the addition.
        """
        if not isinstance(other, Datablock):
            raise ValueError("Can only add Datablock to Datablock")

        new_metadata = {}
        if "ref" in self.metadata:
            new_metadata["ref"] = self.metadata["ref"]
        if "ref" in other.metadata:
            if "ref" in new_metadata:
                new_metadata["ref"] += ", " + other.metadata["ref"]
            else:
                new_metadata["ref"] = other.metadata["ref"]
        warnings.warn("Metadata and labels are not preserved in Datablock addition")

        new_x = np.concatenate((self.x, other.x))
        new_y = np.concatenate((self.y, other.y))
        new_cov_stat = sp.linalg.block_diag(self.cov_stat, other.cov_stat)
        new_cov_syst = sp.linalg.block_diag(self.cov_syst, other.cov_syst)
        return Datablock(
            new_x,
            new_y,
            "x",
            "y",
            cov_stat=new_cov_stat,
            cov_syst=new_cov_syst,
            metadata=new_metadata,
        )

    def __radd__(self, other):
        if other == 0:
            return self
        return self.__add__(other)


class Datawriter:
    r"""
    A class for writing Datablock data to files.

    Parameters
    ----------
    datablock : Datablock
        The Datablock object containing the data to be written.
    file_prefix : str
        The prefix for the output files.
    folder_path : str, optional
        The folder path where the files will be saved. Default is the current directory.
        If the folder does not exist, it will be created.
    quiet : bool, optional
        If True, suppresses output messages. Default is False.
    format : dict, optional
        A dictionary specifying the format for saving the data.
    """

    def __init__(
        self,
        datablock,
        file_prefix,
        folder_path="",
        quiet=False,
        format=dict(fmt="% .11e", delimiter="\t"),
    ):
        self.datablock = datablock
        self.folder = folder_path
        self.file_prefix = file_prefix
        self.header = ""
        self.footer = ""
        self.quiet = quiet
        self.format = format
        if "header" in self.datablock.metadata:
            self.header += self.datablock.metadata["header"]
        if "footer" in self.datablock.metadata:
            self.footer += self.datablock.metadata["footer"]
        os.makedirs(self.folder, exist_ok=True)

    def write_data(self, **filekwargs):
        r"""
        Write the data to a .dat file.
        The file will contain columns for x, y, statistical uncertainty, and systematic uncertainty.

        Parameters
        ----------
        filekwargs : dict
            Additional keyword arguments to be passed to np.savetxt.

        Returns
        -------
        None
        """
        data = np.column_stack(
            (
                self.datablock.x,
                self.datablock.y,
                np.sqrt(np.diag(self.datablock.cov_stat)),
                np.sqrt(np.diag(self.datablock.cov_syst)),
            )
        )
        np.savetxt(
            os.path.join(self.folder, f"{self.file_prefix}.dat"),
            data,
            header=self.header
            + "\n"
            + f"Columns {data.shape}: {self.datablock.xlabel}, {self.datablock.ylabel}, stat_unc, syst_unc",
            footer=self.footer,
            **self.format,
            **filekwargs,
        )
        if not self.quiet:
            print(
                f'"{os.path.join(self.folder, f"{self.file_prefix}.dat")}" created {data.shape}.'
            )

    def write_cov_stat(self, **filekwargs):
        r"""
        Write the statistical covariance matrix to a .dat file.
        The file will contain the statistical covariance matrix.

        Parameters
        ----------
        filekwargs : dict
            Additional keyword arguments to be passed to np.savetxt.

        Returns
        -------
        None
        """
        np.savetxt(
            os.path.join(self.folder, f"{self.file_prefix}_COV_STAT.dat"),
            self.datablock.cov_stat,
            header=self.header
            + "\n"
            + f"Statistical covariance matrix {self.datablock.cov_stat.shape}",
            footer=self.footer,
            **self.format,
            **filekwargs,
        )
        if not self.quiet:
            print(
                f'"{os.path.join(self.folder, f"{self.file_prefix}_COV_STAT.dat")}" created {self.datablock.cov_stat.shape}.'
            )

    def write_cov_syst(self, **filekwargs):
        r"""
        Write the systematic covariance matrix to a .dat file.
        The file will contain the systematic covariance matrix.

        Parameters
        ----------
        filekwargs : dict
            Additional keyword arguments to be passed to np.savetxt.

        Returns
        -------
        None
        """
        np.savetxt(
            os.path.join(self.folder, f"{self.file_prefix}_COV_SYST.dat"),
            self.datablock.cov_syst,
            header=self.header
            + "\n"
            + f"Systematic covariance matrix {self.datablock.cov_syst.shape}",
            footer=self.footer,
            **self.format,
            **filekwargs,
        )
        if not self.quiet:
            print(
                f'"{os.path.join(self.folder, f"{self.file_prefix}_COV_SYST.dat")}" created {self.datablock.cov_syst.shape}.'
            )

    def write_all(self, **filekwargs):
        r"""
        Write all data to .dat files.
        This includes the main data file, the statistical covariance matrix file, and the systematic covariance matrix file.

        Parameters
        ----------
        filekwargs : dict
            Additional keyword arguments to be passed to np.savetxt.
            This argument is shared between all three writing functions.
            For a finer control, call the individual writing functions instead.

        Returns
        -------
        None
        """
        if not self.quiet:
            print(f"- {self.file_prefix} - - - - - - - - - - - - - -")
        self.write_data(**filekwargs)
        self.write_cov_stat(**filekwargs)
        self.write_cov_syst(**filekwargs)


class Datareader:
    r"""
    A class for reading Datablock data from files.

    Parameters
    ----------
    folder : str
        The folder path where the files are located.
    prefix : str
        The prefix for the input files.
    """

    def __init__(self, folder, prefix):
        self.folder = folder
        self.prefix = prefix
        self.file_data = os.path.join(folder, f"{prefix}.dat")
        self.file_cov_stat = os.path.join(folder, f"{prefix}_COV_STAT.dat")
        self.file_cov_syst = os.path.join(folder, f"{prefix}_COV_SYST.dat")

    def read_all(self, **filekwargs):
        r"""
        Read all data from .dat files and return a Datablock object.
        This includes the main data file, the statistical covariance matrix file, and the systematic covariance matrix file.

        Parameters
        ----------
        filekwargs : dict
            Additional keyword arguments to be passed to np.loadtxt.

        Returns
        -------
        Datablock
            A Datablock object containing the data read from the files.

        Raises
        ------
        ValueError
            If the shapes of the covariance matrices do not match the expected dimensions.
            If the uncertainties (squared) do not match the diagonals of the covariance matrices.
        """
        # Load data from file
        data = np.loadtxt(self.file_data, **filekwargs)

        # Load comments from file
        comments = []
        with open(self.file_data, encoding="utf-8") as f:
            for line in f:
                if line.lstrip().startswith("#"):
                    comments.append(line.lstrip()[1:].lstrip().rstrip("\n"))

        # Fill labels and metadata from comments
        xlabel = "x"
        ylabel = "y"
        metadata = {}
        try:
            metadata["ref"] = uncite(comments[0])
        except:
            warnings.warn("Could not extract reference from data file.")
        try:
            columns = ": ".join(comments[-1].split(": ")[1:]).split(", ")
            xlabel = columns[0]
            ylabel = columns[1]
        except:
            warnings.warn("Could not extract column names from data file.")

        # Prepare data and create Datablock
        only_one_datapoint = len(data.shape) == 1
        if only_one_datapoint:
            data = data.reshape(1, -1)
        x = data[..., 0]
        y = data[..., 1]
        stat_unc = data[..., 2]
        syst_unc = data[..., 3]
        cov_stat = np.loadtxt(self.file_cov_stat, **filekwargs)
        cov_syst = np.loadtxt(self.file_cov_syst, **filekwargs)
        if only_one_datapoint:
            cov_stat = cov_stat.reshape((1, 1))
            cov_syst = cov_syst.reshape((1, 1))
        if cov_stat.shape != (len(x), len(x)):
            raise ValueError(
                f"Statistical covariance matrix must be of shape {(len(x), len(x))}, but got {cov_stat.shape}"
            )
        if cov_syst.shape != (len(x), len(x)):
            raise ValueError(
                f"Systematic covariance matrix must be of shape {(len(x), len(x))}, but got {cov_syst.shape}"
            )
        if not np.allclose(stat_unc, np.sqrt(np.diag(cov_stat))):
            raise ValueError(
                "Statistical uncertainties (squared) do not match the diagonal of the statistical covariance matrix"
            )
        if not np.allclose(syst_unc, np.sqrt(np.diag(cov_syst))):
            raise ValueError(
                "Systematic uncertainties (squared) do not match the diagonal of the systematic covariance matrix"
            )
        return Datablock(
            x,
            y,
            xlabel,
            ylabel,
            cov_stat=cov_stat,
            cov_syst=cov_syst,
            metadata=metadata,
        )
